# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""This module contains code related to model card operations"""
from __future__ import absolute_import, print_function

import json
import logging
from datetime import datetime
from typing import Optional, Union, List
from botocore.exceptions import ClientError
from boto3.session import Session as boto3_Session
from six.moves.urllib.parse import urlparse
from sagemaker.session import Session

from sagemaker.model_card.schema_constraints import (
    ModelCardStatusEnum,
    RiskRatingEnum,
    ObjectiveFunctionEnum,
    FacetEnum,
    MetricTypeEnum,
    ENVIRONMENT_CONTAINER_IMAGES_MAX_SIZE,
    MODEL_ARTIFACT_MAX_SIZE,
    METRIC_VALUE_TYPE_MAP,
    TRAINING_DATASETS_MAX_SIZE,
    TRAINING_METRICS_MAX_SIZE,
    USER_PROVIDED_TRAINING_METRICS_MAX_SIZE,
    EVALUATION_DATASETS_MAX_SIZE,
)
from sagemaker.model_card.helpers import (
    _OneOf,
    _IsList,
    _IsModelCardObject,
    _JSONEncoder,
    _DefaultToRequestDict,
    _DefaultFromDict,
    _hash_content_str,
    _read_s3_json,
)
from sagemaker.model_card.evaluation_metric_parsers import (
    EvaluationMetricTypeEnum,
    EVALUATION_METRIC_PARSERS,
)
from sagemaker.utils import _start_waiting, unique_name_from_base


logger = logging.getLogger(__name__)


class Environment(_DefaultToRequestDict, _DefaultFromDict):
    """Training/inference environment."""

    container_image = _IsList(str, ENVIRONMENT_CONTAINER_IMAGES_MAX_SIZE)

    def __init__(self, container_image: List[str]):
        """Initialize an Environment object.

        Args:
            container_image (list[str]): A list of SageMaker training/inference image URIs. The maximum list length is 15.
        """  # noqa E501 # pylint: disable=line-too-long
        self.container_image = container_image


class ModelOverview(_DefaultToRequestDict, _DefaultFromDict):
    """An overview of the model."""

    model_artifact = _IsList(str, MODEL_ARTIFACT_MAX_SIZE)
    inference_environment = _IsModelCardObject(Environment)

    def __init__(
        self,
        model_id: Optional[str] = None,
        model_name: Optional[str] = None,
        model_description: Optional[str] = None,
        model_version: Optional[Union[int, float]] = None,
        problem_type: Optional[str] = None,
        algorithm_type: Optional[str] = None,
        model_creator: Optional[str] = None,
        model_owner: Optional[str] = None,
        model_artifact: Optional[List[str]] = None,
        inference_environment: Optional[Environment] = None,
    ):
        """Initialize a Model Overview object.

        Args:
            model_id (str, optional): A SageMaker Model ARN or non-SageMaker Model ID (default: None).
            model_name (str, optional): A unique name for the model (default: None).
            model_description (str, optional): A description of the model (default: None).
            model_version (int or float, optional): The version of the model (default: None).
            problem_type (str, optional): The type of problem that the model solves. For example, "Binary Classification", "Multiclass Classification", "Linear Regression", "Computer Vision", or "Natural Language Processing" (default: None).
            algorithm_type (str, optional): The algorithm used to solve the problem type (default: None).
            model_creator (str, optional): The organization, research group, or authors that created the model (default: None).
            model_owner (str, optional): The individual or group that maintains the model in your organization (default: None).
            model_artifact (List[str], optional): A list of model artifact location URIs. The maximum list size is 15. (default: None).
            inference_environment (Environment, optional): An overview of the model's inference environment (default: None).
        """  # noqa E501 # pylint: disable=line-too-long
        self.model_id = model_id
        self.model_name = model_name
        self.model_description = model_description
        self.model_version = model_version
        self.problem_type = problem_type
        self.algorithm_type = algorithm_type
        self.model_creator = model_creator
        self.model_owner = model_owner
        self.model_artifact = model_artifact
        self.inference_environment = inference_environment

    @classmethod
    def from_model_name(cls, model_name: str, sagemaker_session: Session = None, **kwargs):
        """Initialize a model overview object from auto-discovered data.

        Args:
            model_name (str): The unique name of the model.
            sagemaker_session (Session, optional): A SageMaker Session
                object, used for SageMaker interactions (default: None). If not
                specified, a SageMaker Session is created using the default AWS configuration
                chain.
            **kwargs: Other arguments in ModelOverview, i.e. model_description,
                problem_type, algorithm_type, model_creator, model_owner, model_version
        Raises:
            ValueError: A model with this name does not exist.
            ValueError: A model card already exists for this model.
        """

        def call_describe_model():
            """Load existing model."""
            try:
                model_response = sagemaker_session.sagemaker_client.describe_model(
                    ModelName=model_name
                )
            except ClientError as e:
                if e.response["Error"]["Message"].startswith(  # pylint: disable=r1720
                    "Could not find model"
                ):
                    raise ValueError(
                        (
                            f"Model details for model {model_name} could not be found. "
                            "Make sure the model name is valid."
                        )
                    )
                else:
                    raise
            return model_response

        def search_model_associated_model_cards(model_id: str):
            """Check if a model card already exists for this model.

            Args:
                model_id (str): A SageMaker model ID.
            """
            response = sagemaker_session.sagemaker_client.search(
                Resource="ModelCard",
                SearchExpression={
                    "Filters": [
                        {
                            "Name": "ModelId",
                            "Operator": "Equals",
                            "Value": model_id,
                        }
                    ]
                },
            )
            return [c["ModelCard"]["ModelCardName"] for c in response["Results"]]

        if not sagemaker_session:
            sagemaker_session = Session()  # pylint: disable=W0106

        model_response = call_describe_model()

        associated_model_cards = search_model_associated_model_cards(model_response["ModelArn"])
        if associated_model_cards:
            raise ValueError(
                f"The model has been associated with {associated_model_cards} model cards."
            )

        if "Containers" in model_response:  # inference pipeline model
            artifacts = [c["ModelDataUrl"] for c in model_response["Containers"]]
        elif (
            "PrimaryContainer" in model_response
            and "ModelDataUrl" in model_response["PrimaryContainer"]
        ):
            artifacts = [model_response["PrimaryContainer"]["ModelDataUrl"]]
        else:
            artifacts = []

        kwargs.update(
            {
                "model_name": model_name,
                "model_id": model_response["ModelArn"],
                "inference_environment": Environment(
                    container_image=[model_response["PrimaryContainer"]["Image"]]
                ),
                "model_artifact": artifacts,
            }
        )

        return cls(**kwargs)


class IntendedUses(_DefaultToRequestDict, _DefaultFromDict):
    """The intended uses of a model."""

    risk_rating = _OneOf(RiskRatingEnum)

    def __init__(
        self,
        purpose_of_model: Optional[str] = None,
        intended_uses: Optional[str] = None,
        factors_affecting_model_efficiency: Optional[str] = None,
        risk_rating: Optional[Union[RiskRatingEnum, str]] = RiskRatingEnum.UNKNOWN,
        explanations_for_risk_rating: Optional[str] = None,
    ):
        """Initialize an Intended Uses object.

        Args:
            purpose_of_model (str, optional): The general purpose of this model (default: None).
            intended_uses (str, optional): The intended use cases for this model (default: None).
            factors_affecting_model_efficiency (str, optional): Factors affecting model efficacy (default: None).
            risk_rating (RiskRatingEnum or str, optional): Your organization's risk rating for this model. It is highly recommended to use sagemaker.model_card.RiskRatingEnum. Possible values include: ``RiskRatingEnum.HIGH`` ("High"), ``RiskRatingEnum.LOW`` ("Low"), ``RiskRatingEnum.MEDIUM`` ("Medium"), or ``RiskRatingEnum.UNKNOWN`` ("Unknown"). Defaults to ``RiskRatingEnum.UNKNOWN``.
            explanations_for_risk_rating (str, optional): An explanation of why your organization categorizes this model with this risk rating (default: None).
        """  # noqa E501 # pylint: disable=line-too-long
        self.purpose_of_model = purpose_of_model
        self.intended_uses = intended_uses
        self.factors_affecting_model_efficiency = factors_affecting_model_efficiency
        self.risk_rating = risk_rating
        self.explanations_for_risk_rating = explanations_for_risk_rating


class Function(_DefaultToRequestDict, _DefaultFromDict):
    """Function details."""

    function = _OneOf(ObjectiveFunctionEnum)
    facet = _OneOf(FacetEnum)

    def __init__(
        self,
        function: Optional[Union[ObjectiveFunctionEnum, str]] = None,
        facet: Optional[Union[FacetEnum, str]] = None,
        condition: Optional[str] = None,
    ):
        """Initialize a Function object.

        Args:
            function (ObjectiveFunctionEnum or str, optional): The optimization direction of the model's objective function. It is highly recommended to use sagemaker.model_card.ObjectiveFunctionEnum. Possible values include: ``ObjectiveFunctionEnum.MAXIMIZE`` ("Maximize") or ``ObjectiveFunctionEnum.MINIMIZE`` ("Minimize") (default: None).
            facet (FacetEnum or str, optional): The metric of the model's objective function. For example, `loss` or `rmse`. It is highly recommended to use sagemaker.model_card.FacetEnum. Possible values include:, ``FacetEnum.ACCURACY`` ("Accuracy"), ``FacetEnum.AUC`` ("AUC"), ``FacetEnum.LOSS`` ("Loss"), ``FacetEnum.MAE`` ("MAE"), or ``FacetEnum.RMSE`` ("RMSE") (default: None).
            condition (str, optional): An optional description of any conditions of your objective function metric (default: None).
        """  # noqa E501 # pylint: disable=line-too-long
        self.function = function
        self.facet = facet
        self.condition = condition


class ObjectiveFunction(_DefaultToRequestDict, _DefaultFromDict):
    """The objective function that is optimized during model training."""

    function = _IsModelCardObject(Function)

    def __init__(
        self,
        function: Function,
        notes: Optional[str] = None,
    ):
        """Initialize an Objective Function object.

        Args:
            function (Function): A Function object that details optimization direction, metric, and additional descriptions.
            notes (str, optional): Notes about the objective function, including other considerations for possible objective functions (default: None).
        """  # noqa E501 # pylint: disable=line-too-long
        self.function = function
        self.notes = notes


class Metric(_DefaultToRequestDict, _DefaultFromDict):
    """Metric data."""

    type = _OneOf(MetricTypeEnum)

    def __init__(
        self,
        name: str,
        type: Union[MetricTypeEnum, str],  # pylint: disable=W0622
        value: Union[int, float, str, bool, List],
        notes: Optional[str] = None,
        x_axis_name: Optional[Union[str, list]] = None,
        y_axis_name: Optional[Union[str, list]] = None,
    ):
        """Initialize a Metric object.

        Args:
            name (str): The name of the metric.
            type (str or MetricTypeEnum): It is highly recommended to use sagemaker.model_card.MetricTypeEnum. Possible values include: ``MetricTypeEnum.BAR_CHART`` ("bar_char"), ``MetricTypeEnum.BOOLEAN`` ("boolean"), ``MetricTypeEnum.LINEAR_GRAPH`` ("linear_graph"), ``MetricTypeEnum.MATRIX`` ("matrix"), ``MetricTypeEnum.NUMBER`` ("number"), or ``MetricTypeEnum.STRING`` ("string").
            value (int or float or str or bool or List): The datatype of the metric. The metric's `value` must be compatible with the metric's `type`.
            notes (str, optional): Any notes to add to the metric (default: None).
            x_axis_name (str, optional): The name of the x axis (default: None).
            y_axis_name (str, optional): The name of the y axis (default: None).
        """  # noqa E501  # pylint: disable=line-too-long
        self.name = name
        self.type = type
        self.value = value
        self.notes = notes
        self.x_axis_name = x_axis_name
        self.y_axis_name = y_axis_name

    def __str__(self):
        """Return str(self)."""
        return f"Metric(name={self.name}, type={self.type}, value={self.value})"

    def __repr__(self):
        """Return repr(self)."""
        return self.__str__()

    @property
    def value(self):
        """Getter for value."""
        return self._value

    @value.setter
    def value(self, val: Union[int, float, str, bool, List]):
        """Setter for value.

        Args:
            val (int or float or str or bool): The metric value.
        Raises:
            ValueError: If the metric `type` doesn't match the metric `value`.
        """
        if type(val) not in METRIC_VALUE_TYPE_MAP[self.type]:
            raise ValueError(
                f"One of type {METRIC_VALUE_TYPE_MAP[self.type]} is expected for metric type {self.type}"
            )
        self._value = val


class TrainingMetric(_DefaultToRequestDict, _DefaultFromDict):
    """Training metric data.

    Should only be used during auto-population of training details.
    """

    def __init__(
        self,
        name: str,
        value: Union[int, float],
        notes: Optional[str] = None,
    ):
        """Initialize a TrainingMetric object.

        Args:
            name (str): The metric name.
            value (int or float): The metric value.
            notes (str, optional): Notes on the metric (default: None).
        """
        self.name = name
        self.value = value
        self.notes = notes


class TrainingJobDetails(_DefaultToRequestDict, _DefaultFromDict):
    """The overview of a training job."""

    training_datasets = _IsList(str, TRAINING_DATASETS_MAX_SIZE)
    training_metrics = _IsList(TrainingMetric, TRAINING_METRICS_MAX_SIZE)
    user_provided_training_metrics = _IsList(
        TrainingMetric, USER_PROVIDED_TRAINING_METRICS_MAX_SIZE
    )
    training_environment = _IsModelCardObject(Environment)

    def __init__(
        self,
        training_arn: Optional[str] = None,
        training_datasets: Optional[List[str]] = None,
        training_environment: Optional[Environment] = None,
        training_metrics: Optional[List[TrainingMetric]] = None,
        user_provided_training_metrics: Optional[List[TrainingMetric]] = None,
    ):
        """Initialize a Training Job Details object.

        Args:
            training_arn (str, optional): The SageMaker training job Amazon Resource Name (ARN) (default: None).
            training_datasets (List[str], optional): The location of the datasets used to train the model. The maximum list size is 15. (default: None).
            training_environment (Environment, optional): The SageMaker training image URI. (default: None).
            training_metrics (list[TrainingMetric], optional): SageMaker training job results. The maximum `training_metrics` list length is 50 (default: None).
            user_provided_training_metrics (list[TrainingMetric], optional): Custom training job results. The maximum `user_provided_training_metrics` list length is 50 (default: None).
        """  # noqa E501 # pylint: disable=line-too-long
        self.training_arn = training_arn
        self.training_datasets = training_datasets
        self.training_environment = training_environment
        self.training_metrics = training_metrics
        self.user_provided_training_metrics = user_provided_training_metrics


class TrainingDetails(_DefaultToRequestDict, _DefaultFromDict):
    """The overview of model training."""

    objective_function = _IsModelCardObject(ObjectiveFunction)
    training_job_details = _IsModelCardObject(TrainingJobDetails)

    def __init__(
        self,
        objective_function: Optional[ObjectiveFunction] = None,
        training_observations: Optional[str] = None,
        training_job_details: Optional[TrainingJobDetails] = None,
    ):
        """Initialize a TrainingDetails object.

        Args:
            objective_function (ObjectiveFunction, optional): The objective function that is optimized during training (default: None).
            training_observations (str, optional): Any observations about training (default: None).
            training_job_details (TrainingJobDetails, optional): Details about any associated training jobs (default: None).
        """  # noqa E501 # pylint: disable=line-too-long
        self.objective_function = objective_function
        self.training_observations = training_observations
        self.training_job_details = training_job_details

    @staticmethod
    def _create_training_details(training_job_data: dict, cls: "TrainingDetails", **kwargs):
        """Create a Training Details object from the data queried from APIs.

        Args:
            training_job_data (dict): Training job data queried from either
                search or `describe_training_job`.
            cls (TrainingDetails): TrainingDetails class.
        """

        if training_job_data:
            job = {
                "training_arn": training_job_data["TrainingJobArn"],
                "training_environment": Environment(
                    container_image=[training_job_data["AlgorithmSpecification"]["TrainingImage"]]
                ),
                "training_metrics": [
                    TrainingMetric(i["MetricName"], i["Value"])
                    for i in training_job_data["FinalMetricDataList"]
                ]
                if "FinalMetricDataList" in training_job_data
                else [],
            }
            kwargs.update({"training_job_details": TrainingJobDetails(**job)})
            instance = cls(**kwargs)
        else:
            instance = None

        return instance

    @classmethod
    def from_model_overview(
        cls, model_overview: ModelOverview, sagemaker_session: Session = None, **kwargs
    ):
        """Initialize a Training Details object from an auto-discovered model overview.

        Args:
            model_overview (ModelOverview): A ModelOverview object.
            sagemaker_session (Session, optional): A SageMaker Session object, used for SageMaker interactions (default: None). If not specified, a SageMaker Session is created using the default AWS configuration chain.
            **kwargs: Other arguments in TrainingDetails, i.e. objective_function, training_observations
        """  # noqa E501 # pylint: disable=line-too-long

        def call_search_training_job():
            """Search training job."""
            res = sagemaker_session.sagemaker_client.search(
                Resource="TrainingJob",
                SearchExpression={
                    "Filters": [
                        {
                            "Name": "ModelArtifacts.S3ModelArtifacts",
                            "Operator": "Equals",
                            "Value": model_overview.model_artifact[0],
                        }
                    ]
                },
            )

            training_job_data = None
            if len(res["Results"]) == 1:
                training_job_data = res["Results"][0]["TrainingJob"]

            return training_job_data

        if not sagemaker_session:
            sagemaker_session = Session()  # pylint: disable=W0106

        training_job_data = None
        if len(model_overview.model_artifact) == 1:  # unique training job name
            training_job_data = call_search_training_job()
        elif len(model_overview.model_artifact) == 0:
            logger.warning(
                (
                    "TraininigJobDetails auto-discovery failed. "
                    "No associated training job. "
                    "Please create one from scrach with TrainingJobDetails "
                    "or use from_training_job_name() instead."
                )
            )
        else:
            logger.warning(
                (
                    "TraininigJobDetails auto-discovery failed. "
                    "There are %s associated training jobs. "
                    "Further clarification is required. "
                    "You could use TrainingDetails.training_job_name after "
                    "which training job to use is decided."
                ),
                len(model_overview.model_artifact),
            )
        return cls._create_training_details(training_job_data=training_job_data, cls=cls, **kwargs)

    @classmethod
    def from_training_job_name(
        cls, training_job_name: str, sagemaker_session: Session = None, **kwargs
    ):
        """Initialize a Training Details object from a training job name.

        Args:
            training_job_name (str): Training job name.
            sagemaker_session (Session, optional): A SageMaker Session
                object, used for SageMaker interactions (default: None). If not
                specified, a SageMaker Session is created using the default AWS configuration
                chain.
            **kwargs: Other arguments in TrainingDetails, i.e. objective_function,
                training_observations
        Raises:
            ValueError: No traininig job is found.
            ValueError: multiple training jobs are matched.
        """

        def call_describe_training_job():
            """Load existing training job."""
            try:
                training_job_data = sagemaker_session.sagemaker_client.describe_training_job(
                    TrainingJobName=training_job_name
                )
            except ClientError as e:
                if (  # pylint: disable=r1720
                    e.response["Error"]["Message"] == "Requested resource not found."
                ):
                    raise ValueError(
                        (
                            "Training job details could not be found. "
                            "Make sure the training job name is valid."
                        )
                    )
                else:
                    raise

            return training_job_data

        if not sagemaker_session:
            sagemaker_session = Session()  # pylint: disable=W0106

        training_job_data = call_describe_training_job()

        return cls._create_training_details(training_job_data=training_job_data, cls=cls, **kwargs)

    def add_metric(self, metric: TrainingMetric):
        """Add custom training metrics.

        Args:
            metric (TrainingMetric): The custom metric to add.
        """
        if not self.training_job_details:
            self.training_job_details = TrainingJobDetails()
        self.training_job_details.user_provided_training_metrics.append(metric)


class MetricGroup(_DefaultToRequestDict, _DefaultFromDict):
    """Group of metric data"""

    metric_data = _IsList(Metric)

    def __init__(self, name: str, metric_data: Optional[List[Metric]] = None):
        """Initialize a Metric Group object.

        Args:
            name (str): The metric group name.
            metric_data (List[Metric]): A list of Metric objects.
        """
        self.name = name
        self.metric_data = metric_data

    def __str__(self):
        """Return str(self)."""
        metric_size = 0 if self.metric_data is None else len(self.metric_data)
        return f"MetricGroup(name={self.name}, size={metric_size})"

    def __repr__(self):
        """Return repr(self)."""
        return self.__str__()

    def add_metric(self, metric: Metric):
        """Add metric to the metric group.

        Args:
            metric (Metric): The Metric object to add.
        """
        self.metric_data.append(metric)


class EvaluationJob(_DefaultToRequestDict, _DefaultFromDict):
    """Overview of an evaluation job."""

    datasets = _IsList(str, EVALUATION_DATASETS_MAX_SIZE)
    metric_groups = _IsList(MetricGroup)

    def __init__(
        self,
        name: str,
        evaluation_observation: Optional[str] = None,
        evaluation_job_arn: Optional[str] = None,
        datasets: Optional[List[str]] = None,
        metadata: Optional[dict] = None,
        metric_groups: Optional[List[MetricGroup]] = None,
    ):
        """Initialize an Evaluation Job object.

        Args:
            name (str): The evaluation job name.
            evaluation_observation (str, optional): Any observations made during model evaluation (default: None).
            evaluation_job_arn (str, optional): The Amazon Resource Name (ARN) of the evaluation job (default: None).
            datasets (List[str], optional): Evaluation dataset locations. Maximum list length is 10 (default: None).
            metadata (Optional[dict], optional): Additional attributes associated with the evaluation results (default: None).
            metric_groups (List[MetricGroup], optional): An evaluation Metric Group object (default: None).
        """  # noqa E501 # pylint: disable=line-too-long
        self.name = name
        self.evaluation_observation = evaluation_observation
        self.evaluation_job_arn = evaluation_job_arn
        self.datasets = datasets
        self.metadata = metadata
        self.metric_groups = metric_groups

    def get_metric_group(self, group_name):
        """Get a metric group.

        Args:
            group_name (str): The metric group name.
        """
        return self.metric_groups.to_map(key_attribute="name").get(group_name)

    def add_metric_group(self, group_name: str):
        """Add a metric group

        Args:
            group_name (str): The metric group name.
        """
        metric_group = MetricGroup(name=group_name)
        self.metric_groups.append(metric_group)
        return metric_group

    def _add_metric_groups_data(self, metric_groups: List):
        """Encode the model card evaluation metrics to Metric and MetricGroup.

        Args:
            metric_groups (List): List of metric groups raw data.
        """
        for group_data in metric_groups:
            group = self.add_metric_group(group_data["name"])
            for item in group_data["metric_data"]:
                group.add_metric(Metric(**item))

    def _parse_evaluation_metric_json(self, json_data: dict, metric_type: EvaluationMetricTypeEnum):
        """Parse the evaluation metric data to the model card metric schema.

        Args:
            json_data (dict): Metric JSON data.
            metric_type EvaluationMetricTypeEnum: The evaluation metric type
                for the data in the evaluation metrics JSON file.
                Possible values include: ``EvaluationMetricTypeEnum.MODEL_CARD_METRIC_SCHEMA``, ``EvaluationMetricTypeEnum.CLARIFY_BIAS``,
                ``EvaluationMetricTypeEnum.CLARIFY_EXPLAINABILITY``, ``EvaluationMetricTypeEnum.REGRESSION``, ``EvaluationMetricTypeEnum.BINARY_CLASSIFICATION``,
                or ``EvaluationMetricTypeEnum.MULTICLASS_CLASSIFICATION``.
        """  # noqa E501 # pylint: disable=line-too-long
        if not isinstance(metric_type, EvaluationMetricTypeEnum):
            raise ValueError("Please use sagemaker.model_card.EvaluationMetricTypeEnum")

        parser = EVALUATION_METRIC_PARSERS[metric_type]
        return parser.run(json_data)

    def add_metric_group_from_json(
        self,
        json_path: str,
        metric_type: EvaluationMetricTypeEnum = EvaluationMetricTypeEnum.MODEL_CARD_METRIC_SCHEMA,
    ):
        """Add evaluation metric files from S3 bucket.

        Args:
            json_path (str): The path for the evaluation metrics JSON file.
            metric_type (EvaluationMetricTypeEnum, optional): The evaluation metric type
                for the data in the evaluation metrics JSON file.
                Possible values include:
                ``EvaluationMetricTypeEnum.MODEL_CARD_METRIC_SCHEMA``,
                ``EvaluationMetricTypeEnum.CLARIFY_BIAS``,
                ``EvaluationMetricTypeEnum.CLARIFY_EXPLAINABILITY``,
                ``EvaluationMetricTypeEnum.REGRESSION``,
                ``EvaluationMetricTypeEnum.BINARY_CLASSIFICATION``,
                or `EvaluationMetricTypeEnum.MULTICLASS_CLASSIFICATION``
                (default: ``EvaluationMetricTypeEnum.MODEL_CARD_METRIC_SCHEMA``).
        """
        with open(json_path, "r", encoding="utf-8") as istr:
            json_data = json.load(istr)

        result = self._parse_evaluation_metric_json(json_data, metric_type)
        self._add_metric_groups_data(result["metric_groups"])

    def add_metric_group_from_s3(
        self,
        session: boto3_Session,
        s3_url: str,
        metric_type: EvaluationMetricTypeEnum = EvaluationMetricTypeEnum.MODEL_CARD_METRIC_SCHEMA,
    ):
        """Add evaluation metric files from an S3 bucket.

        Args:
            session (Session): A Boto3 session.
            s3_url (str): The S3 URL for the evaluation metrics JSON file.
            metric_type (EvaluationMetricTypeEnum, optional): The evaluation metric type
                for the data in the evaluation metrics JSON file.
                Possible values include:
                ``EvaluationMetricTypeEnum.MODEL_CARD_METRIC_SCHEMA``,
                ``EvaluationMetricTypeEnum.CLARIFY_BIAS``,
                ``EvaluationMetricTypeEnum.CLARIFY_EXPLAINABILITY``,
                ``EvaluationMetricTypeEnum.REGRESSION``,
                ``EvaluationMetricTypeEnum.BINARY_CLASSIFICATION``,
                or ``EvaluationMetricTypeEnum.MULTICLASS_CLASSIFICATION``
                (default: ``EvaluationMetricTypeEnum.MODEL_CARD_METRIC_SCHEMA``).
        """
        parsed_url = urlparse(s3_url)
        bucket = parsed_url.netloc
        key = parsed_url.path.lstrip("/")
        json_data = _read_s3_json(session, bucket, key)

        result = self._parse_evaluation_metric_json(json_data, metric_type)
        self._add_metric_groups_data(result["metric_groups"])


class AdditionalInformation(_DefaultToRequestDict, _DefaultFromDict):
    """Additional information for a model card."""

    def __init__(
        self,
        ethical_considerations: Optional[str] = None,
        caveats_and_recommendations: Optional[str] = None,
        custom_details: Optional[dict] = None,
    ):
        """Initialize an Additional Information object.

        Args:
            ethical_considerations (str, optional): Any ethical considerations to document about the model (default: None).
            caveats_and_recommendations (str, optional): Caveats and recommendations for those who might use this model in their applications (default: None).
            custom_details (dict, optional): Any additional custom information to document about the model (default: None).
        """  # noqa E501 # pylint: disable=line-too-long
        self.ethical_considerations = ethical_considerations
        self.caveats_and_recommendations = caveats_and_recommendations
        self.custom_details = custom_details


class ModelCard(object):
    """Use an Amazon SageMaker Model Card to document qualitative and quantitative information about a model."""  # noqa E501  # pylint: disable=c0301

    DECODER_ATTRIBUTE_MAP = {
        "ModelCardName": "name",
        "ModelCardArn": "arn",
        "ModelCardStatus": "status",
        "ModelCardVersion": "version",
        "CreationTime": "created_time",
        "CreatedBy": "created_by",
        "LastModifiedTime": "last_modified_time",
        "LastModifiedBy": "last_modified_by",
    }
    CREATE_MODEL_CARD_REQUIRED = ["ModelCardName", "ModelCardStatus"]
    INITIAL_VERSION = 1

    status = _OneOf(ModelCardStatusEnum)
    model_overview = _IsModelCardObject(ModelOverview)
    intended_uses = _IsModelCardObject(IntendedUses)
    training_details = _IsModelCardObject(TrainingDetails)
    evaluation_details = _IsList(EvaluationJob)
    additional_information = _IsModelCardObject(AdditionalInformation)

    def __init__(
        self,
        name: str,
        status: Optional[Union[ModelCardStatusEnum, str]] = ModelCardStatusEnum.DRAFT,
        arn: Optional[str] = None,
        version: Optional[int] = None,
        created_time: Optional[datetime] = None,
        created_by: Optional[dict] = None,
        last_modified_time: Optional[datetime] = None,
        last_modified_by: Optional[dict] = None,
        model_overview: Optional[ModelOverview] = None,
        intended_uses: Optional[IntendedUses] = None,
        training_details: Optional[TrainingDetails] = None,
        evaluation_details: Optional[List[EvaluationJob]] = None,
        additional_information: Optional[AdditionalInformation] = None,
        sagemaker_session: Optional[Session] = None,
    ):
        """Initialize an Amazon SageMaker Model Card.

        Args:
            name (str): The unique name of the model card.
            status (ModelCardStatusEnum or str, optional): Your organization's approval status of the model card. It is highly recommended to use sagemaker.model_card.ModelCardStatusEnum. Possible values include: ``ModelCardStatusEnum.APPROVED`` ("Approved"), ``ModelCardStatusEnum.ARCHIVED`` ("Archived"), ``ModelCardStatusEnum.DRAFT`` ("Draft"), or ``ModelCardStatusEnum.PENDING_REVIEW`` ("PendingReview"). Defaults to ``ModelCardStatusEnum.DRAFT``.
            arn (str, optional): The Amazon Resource Name (ARN) of the model card (default: None).
            version (int, optional): The model card version (default: None).
            created_time (datetime, optional): The date/time that you created the model card (default: None).
            created_by (dict, optional): The group or individual that created the model card (default: None).
            last_modified_time (datetime, optional): The last time that the model card was modified (default: None).
            last_modified_by (dict, optional): The group or individual that last modified the model card (default: None).
            model_overview (ModelOverview, optional): An overview of the model (default: None).
            intended_uses (IntendedUses, optional): The intended uses of the model (default: None).
            training_details (TrainingDetails, optional): The training details of the model (default: None).
            evaluation_details (List[EvaluationJob], optional): The evaluation details of the model (default: None).
            additional_information (AdditionalInformation, optional): Additional information about the model (default: None).
            sagemaker_session (Session, optional): A SageMaker Session object, used for SageMaker interactions (default: None). If not specified, a SageMaker Session is created using the default AWS configuration chain.
        """  # noqa E501 # pylint: disable=line-too-long
        self.name = name
        self.arn = arn
        self.status = status
        self.version = version
        self.created_time = created_time
        self.created_by = created_by
        self.last_modified_time = last_modified_time
        self.last_modified_by = last_modified_by
        self.model_overview = model_overview
        self.intended_uses = intended_uses
        self.training_details = training_details
        self.evaluation_details = evaluation_details
        self.additional_information = additional_information
        self.sagemaker_session = sagemaker_session or Session()

    def create(self):
        """Create the model card"""

        if self.arn:  # model card has been created
            raise FileExistsError(f"Model card {self.name} already exists with arn {self.arn}")

        request_args = self._create_request_args()
        logger.info("Creating model card with name: %s", self.name)
        logger.debug("CreateModelCard request: %s", json.dumps(request_args, indent=4))
        self.sagemaker_session.sagemaker_client.create_model_card(**request_args)

        # udpate model card with the latest data from server
        response = self.sagemaker_session.sagemaker_client.describe_model_card(
            ModelCardName=self.name
        )
        self.created_time = response["CreationTime"]
        self.last_modified_time = response["LastModifiedTime"]
        self.version = response["ModelCardVersion"]
        self.arn = response["ModelCardArn"]

        return self.arn

    def _create_request_args(self):
        """Generate the request body for create model card call."""
        request_args = {}
        for arg in ModelCard.CREATE_MODEL_CARD_REQUIRED:
            request_args[arg] = getattr(self, ModelCard.DECODER_ATTRIBUTE_MAP[arg])

        content = {}
        for attr_name, attr_val in ModelCard.__dict__.items():
            if type(attr_val) in [_IsModelCardObject, _IsList]:
                object_val = getattr(self, attr_name)
                if object_val:
                    content[attr_name] = object_val

        request_args["Content"] = json.dumps(content, cls=_JSONEncoder)

        return request_args

    @classmethod
    def load(cls, name: str, version: Optional[int] = None, sagemaker_session: Session = None):
        """Load a model card.

        Args:
            name (str): The unique name of the model card.
            version (int, optional): The model card version (default: None).
            sagemaker_session (Session, optional): A SageMaker Session
                object, used for SageMaker interactions (default: None). If not
                specified, a SageMaker Session is created using the default AWS configuration
                chain.
        """

        def decode_attributes(response: dict):
            """Decode attributes from API response to class attributes.

            Args:
                response (dict): describe_model_card response.
            """
            decoded = {}
            for var, attr in cls.DECODER_ATTRIBUTE_MAP.items():
                if var in response:
                    decoded[attr] = response[var]

            content = json.loads(response["Content"])
            decoded = {**decoded, **content}

            return decoded

        if not sagemaker_session:
            sagemaker_session = Session()  # pylint: disable=W0106

        logger.info("Load model card: %s", name)
        request_args = {"ModelCardName": name}
        if version:
            request_args["ModelCardVersion"] = version
        response = sagemaker_session.sagemaker_client.describe_model_card(**request_args)

        model_card_args = {"sagemaker_session": sagemaker_session, **decode_attributes(response)}

        return cls(**model_card_args)

    def update(self, **kwargs):
        """Update the model card.

        Raises:
            ValueError: _description_

        Returns:
            _type_: _description_
        """

        # use created_time to infer if the model card
        if self.created_time is None:
            raise ValueError(
                "Please create a model card or load an existing model card before update."
            )

        logger.info("Updating model card: %s", self.name)

        # update the current attributes if any
        for attribute, value in kwargs.items():
            if hasattr(self, attribute):
                setattr(self, attribute, value)
            else:
                logger.warning("%s doesn't exist in model card.", attribute)

        # Get the latest model card on the server
        previous = self.sagemaker_session.sagemaker_client.describe_model_card(
            ModelCardName=self.name
        )
        previous_content_hash = _hash_content_str(previous["Content"])

        current = self._create_request_args()
        logger.debug("UpdateModelCard request: %s", json.dumps(current, indent=4))

        result = {}
        if previous["ModelCardStatus"] != current["ModelCardStatus"]:
            logger.info("Update model card status")
            update_status_response = self.sagemaker_session.sagemaker_client.update_model_card(
                ModelCardName=current["ModelCardName"], ModelCardStatus=current["ModelCardStatus"]
            )
            result["status"] = update_status_response

        current_content_hash = _hash_content_str(current["Content"])
        if previous_content_hash != current_content_hash:
            logger.info("Update model card content")
            update_content_response = self.sagemaker_session.sagemaker_client.update_model_card(
                ModelCardName=current["ModelCardName"], Content=current["Content"]
            )
            result["content"] = update_content_response

        return result

    def delete(self):
        """Delete the model card"""
        logger.info("Deleting model card with name: %s", self.name)
        self.sagemaker_session.sagemaker_client.delete_model_card(ModelCardName=self.name)
        self.status = ModelCardStatusEnum.ARCHIVED

        return True

    def export_pdf(
        self,
        s3_output_path: str,
        export_job_name: Optional[str] = None,
        model_card_version: Optional[int] = None,
    ):
        """Export the model card as a PDF.

        Args:
            s3_output_path (str): The S3 output path for your model card PDF.
            export_job_name (str, optional): The model card export job name.
                (default: None).
            model_card_version (int, optional): The version of the model card you want to export.
                (default: None).
        """
        if not model_card_version:
            model_card_version = self.version

        if not export_job_name:
            export_job_name = unique_name_from_base(self.name)

        job = ModelCardExportJob(
            export_job_name=export_job_name,
            model_card_name=self.name,
            model_card_version=model_card_version,
            sagemaker_session=self.sagemaker_session,
            s3_output_path=s3_output_path,
        )
        output_path = job.create()
        return output_path

    def list_export_jobs(self, **kwargs):
        """List all the model card export jobs for a specific model card.

        Args:
            kwargs: additional parameters from
                sagemaker_session.sagemaker_client.list_model_card_export_jobs.
        """
        return ModelCardExportJob.list_export_jobs(
            model_card_name=self.name, sagemaker_session=self.sagemaker_session, **kwargs
        )

    def get_version_history(self, **kwargs):
        """Get the version history of the model card.

        Args:
            kwargs: additional parameters from
                sagemaker_session.sagemaker_client.list_model_card_versions.
        """
        response = self.sagemaker_session.sagemaker_client.list_model_card_versions(
            ModelCardName=self.name, **kwargs
        )

        return response["ModelCardVersionSummaryList"]


class ModelCardExportJob(object):
    """Model card export job class."""

    EXPORT_JOB_POLLING_FREQUENCY = 4  # seconds

    def __init__(
        self,
        export_job_name: str,
        model_card_name: str,
        model_card_version: int,
        s3_output_path: str,
        s3_export_artifacts: Optional[str] = None,
        export_job_arn: Optional[str] = None,
        sagemaker_session: Optional[Session] = None,
        status: Optional[str] = None,
        failure_reason: Optional[str] = None,
    ):
        """Initialize a model card export job instance.

        Args:
            export_job_name (str): The model card export job name.
            model_card_name (str): The unique name of the model card you want to export.
            model_card_version (int): The version of the model card you want to export.
            s3_output_path (str): The S3 output path for your model card PDF.
            s3_export_artifacts (str, optional): The full S3 URI for your model card PDF
                (default: None).
            sagemaker_session (Session, optional): A SageMaker Session
                object, used for SageMaker interactions (default: None). If not
                specified, a SageMaker Session is created using the default AWS configuration
                chain.
            status (str, optional): The status of the model card export job.
                (default: None).
            export_job_arn: (str, optional): The Amazon Resource Name (ARN) of the model card export job.
                (default: None).
            failure_reason: (str, optional): The failure reason if your model card fails to export.
                (default: None).
        """  # noqa E501 # pylint: disable=line-too-long
        self.export_job_name = export_job_name
        self.model_card_name = model_card_name
        self.model_card_version = model_card_version
        self.s3_output_path = s3_output_path
        self.s3_export_artifacts = s3_export_artifacts
        self.sagemaker_session = sagemaker_session or Session()
        self.export_job_arn = export_job_arn
        self.status = status
        self.failure_reason = failure_reason

    def create(self):
        """Create a model card export job."""
        output_config = {"S3OutputPath": self.s3_output_path}
        response = self.sagemaker_session.sagemaker_client.create_model_card_export_job(
            ModelCardName=self.model_card_name,
            ModelCardVersion=self.model_card_version,
            ModelCardExportJobName=self.export_job_name,
            OutputConfig=output_config,
        )
        self.export_job_arn = response["ModelCardExportJobArn"]

        # Wait for the job to finish
        job = ModelCardExportJob.load(self.export_job_arn, self.sagemaker_session)
        while job.status == "InProgress":
            job = ModelCardExportJob.load(self.export_job_arn, self.sagemaker_session)
            _start_waiting(self.EXPORT_JOB_POLLING_FREQUENCY)

        if job.status == "Failed":
            logger.warning(
                "Failed to export model card to %s. %s", job.s3_export_artifacts, job.failure_reason
            )
            output = None
        elif job.status == "Completed":
            logger.info(
                "Model card %s is successfully exported to %s.",
                job.model_card_name,
                job.s3_export_artifacts,
            )
            output = job.s3_export_artifacts

        return output

    @classmethod
    def load(cls, export_job_arn: str, sagemaker_session: Session = None):
        """Load the model card export job.

        Args:
            export_job_arn (str): The Amazon Resource Name (ARN) of the export job.
            sagemaker_session (Session, optional): A SageMaker Session
                object, used for SageMaker interactions (default: None). If not
                specified, a SageMaker Session is created using the default AWS configuration
                chain.
        """
        if not sagemaker_session:
            sagemaker_session == Session()  # pylint: disable=W0106

        response = sagemaker_session.sagemaker_client.describe_model_card_export_job(
            ModelCardExportJobArn=export_job_arn
        )

        key_args = {
            "export_job_name": response["ModelCardExportJobName"],
            "model_card_name": response["ModelCardName"],
            "model_card_version": response["ModelCardVersion"],
            "sagemaker_session": sagemaker_session,
            "status": response["Status"],
            "export_job_arn": response["ModelCardExportJobArn"],
            "s3_output_path": response["OutputConfig"]["S3OutputPath"],
        }

        if "ExportArtifacts" in response:
            key_args["s3_export_artifacts"] = response["ExportArtifacts"]["S3ExportArtifacts"]
        if response["Status"] == "Failed" and "FailureReason" in response:
            key_args["failure_reason"] = response["FailureReason"]

        return cls(**key_args)

    @staticmethod
    def list_export_jobs(
        model_card_name: str, sagemaker_session: Optional[Session] = None, **kwargs
    ):
        """List all model card export jobs for a specific model card.

        Args:
            model_card_name (str): The unique name of the model card.
            sagemaker_session (Session, optional): A SageMaker Session
                object, used for SageMaker interactions (default: None). If not
                specified, a SageMaker Session is created using the default AWS configuration
                chain.
            kwargs: additional parameters from
                sagemaker_session.sagemaker_client.list_model_card_export_jobs.
        """
        if not sagemaker_session:
            sagemaker_session = Session()  # pylint: disable=W0106

        return sagemaker_session.sagemaker_client.list_model_card_export_jobs(
            ModelCardName=model_card_name, **kwargs
        )



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Deprecated content type constants. Just use the mime type strings."""
from __future__ import absolute_import

import deprecations

deprecations.removed_warning("The sagemaker.content_types module")

CONTENT_TYPE_JSON = "application/json"
CONTENT_TYPE_CSV = "text/csv"
CONTENT_TYPE_OCTET_STREAM = "application/octet-stream"
CONTENT_TYPE_NPY = "application/x-npy"



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Classes for using model cards with Amazon SageMaker."""

from sagemaker.model_card.model_card import (  # noqa: F401 # pylint: disable=unused-import
    Environment,
    ModelOverview,
    IntendedUses,
    ObjectiveFunction,
    TrainingMetric,
    Metric,
    Function,
    TrainingJobDetails,
    TrainingDetails,
    MetricGroup,
    EvaluationJob,
    AdditionalInformation,
    ModelCard,
)

from sagemaker.model_card.schema_constraints import (  # noqa: F401 # pylint: disable=unused-import
    ModelCardStatusEnum,
    RiskRatingEnum,
    ObjectiveFunctionEnum,
    FacetEnum,
    MetricTypeEnum,
)

from sagemaker.model_card.evaluation_metric_parsers import (  # noqa: F401 # pylint: disable=unused-import
    EvaluationMetricTypeEnum,
)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Contains the SageMaker Experiment Run class."""
from __future__ import absolute_import

import datetime
import logging
from enum import Enum
from math import isnan, isinf
from numbers import Number
from typing import Optional, List, Dict, TYPE_CHECKING, Union

import dateutil
from numpy import array

from sagemaker.apiutils import _utils
from sagemaker.experiments import _api_types
from sagemaker.experiments._api_types import TrialComponentArtifact, _TrialComponentStatusType
from sagemaker.experiments._helper import (
    _ArtifactUploader,
    _LineageArtifactTracker,
)
from sagemaker.experiments._environment import _RunEnvironment
from sagemaker.experiments._run_context import _RunContext
from sagemaker.experiments.experiment import _Experiment
from sagemaker.experiments._metrics import _MetricsManager
from sagemaker.experiments.trial import _Trial
from sagemaker.experiments.trial_component import _TrialComponent

from sagemaker.utils import (
    get_module,
    unique_name_from_base,
)

from sagemaker.experiments._utils import (
    guess_media_type,
    resolve_artifact_name,
    verify_length_of_true_and_predicted,
    validate_invoked_inside_run_context,
    get_tc_and_exp_config_from_job_env,
    verify_load_input_names,
    is_run_trial_component,
)

if TYPE_CHECKING:
    from sagemaker import Session

logger = logging.getLogger(__name__)

RUN_NAME_BASE = "Sagemaker-Run".lower()
TRIAL_NAME_TEMPLATE = "Default-Run-Group-{}"
MAX_RUN_TC_ARTIFACTS_LEN = 30
MAX_NAME_LEN_IN_BACKEND = 120
EXPERIMENT_NAME = "ExperimentName"
TRIAL_NAME = "TrialName"
RUN_NAME = "RunName"
DELIMITER = "-"
RUN_TC_TAG_KEY = "sagemaker:trial-component-source"
RUN_TC_TAG_VALUE = "run"
RUN_TC_TAG = {"Key": RUN_TC_TAG_KEY, "Value": RUN_TC_TAG_VALUE}


class SortByType(Enum):
    """The type of property by which to sort the `list_runs` results."""

    CREATION_TIME = "CreationTime"
    NAME = "Name"


class SortOrderType(Enum):
    """The type of order to sort the list or search results."""

    ASCENDING = "Ascending"
    DESCENDING = "Descending"


class Run(object):
    """A collection of parameters, metrics, and artifacts to create a ML model."""

    def __init__(
        self,
        experiment_name: str,
        run_name: Optional[str] = None,
        experiment_display_name: Optional[str] = None,
        run_display_name: Optional[str] = None,
        tags: Optional[List[Dict[str, str]]] = None,
        sagemaker_session: Optional["Session"] = None,
    ):
        """Construct a `Run` instance.

        SageMaker Experiments automatically tracks the inputs, parameters, configurations,
        and results of your iterations as runs.
        You can assign, group, and organize these runs into experiments.
        You can also create, compare, and evaluate runs.

        The code sample below shows how to initialize a run, log parameters to the Run object
        and invoke a training job under the context of this Run object, which automatically
        passes the run's ``experiment_config`` (including the experiment name, run name etc.)
        to the training job.

        Note:
            All log methods (e.g. ``log_parameter``, ``log_metric``, etc.) have to be called within
            the run context (i.e. the ``with`` statement). Otherwise, a ``RuntimeError`` is thrown.

        .. code:: python

            with Run(experiment_name="my-exp", run_name="my-run", ...) as run:
                run.log_parameter(...)
                ...
                estimator.fit(job_name="my-job")  # Create a training job

        In order to reuse an existing run to log extra data, ``load_run`` is recommended.
        The code snippet below displays how to load the run initialized above
        in a custom training job script, where no ``run_name`` or ``experiment_name``
        is presented as they are automatically retrieved from the experiment config
        in the job environment.

        Note:
            Instead of the ``Run`` constructor, the ``load_run`` is recommended to use
            in a job script to load the existing run created before the job launch.
            Otherwise, a new run may be created each time you launch a job.

        .. code:: python

            with load_run() as run:
                run.log_metric(...)
                ...

        Args:
            experiment_name (str): The name of the experiment. The name must be unique
                within an account.
            run_name (str): The name of the run. If it is not specified, one is auto generated.
            experiment_display_name (str): Name of the experiment that will appear in UI,
                such as SageMaker Studio. (default: None). This display name is used in
                a create experiment call. If an experiment with the specified name already exists,
                this display name won't take effect.
            run_display_name (str): The display name of the run used in UI (default: None).
                This display name is used in a create run call. If a run with the
                specified name already exists, this display name won't take effect.
            tags (List[Dict[str, str]]): A list of tags to be used for all create calls,
                e.g. to create an experiment, a run group, etc. (default: None).
            sagemaker_session (sagemaker.session.Session): Session object which
                manages interactions with Amazon SageMaker APIs and any other
                AWS services needed. If not specified, one is created using the
                default AWS configuration chain.
        """
        # TODO: we should revert the lower casting once backend fix reaches prod
        self.experiment_name = experiment_name.lower()
        sagemaker_session = sagemaker_session or _utils.default_session()
        self.run_name = run_name or unique_name_from_base(RUN_NAME_BASE)

        # avoid confusion due to mis-match in casing between run name and TC name
        self.run_name = self.run_name.lower()

        trial_component_name = Run._generate_trial_component_name(
            run_name=self.run_name, experiment_name=self.experiment_name
        )
        self.run_group_name = Run._generate_trial_name(self.experiment_name)

        self._experiment = _Experiment._load_or_create(
            experiment_name=self.experiment_name,
            display_name=experiment_display_name,
            tags=tags,
            sagemaker_session=sagemaker_session,
        )

        self._trial = _Trial._load_or_create(
            experiment_name=self.experiment_name,
            trial_name=self.run_group_name,
            tags=tags,
            sagemaker_session=sagemaker_session,
        )

        self._trial_component, is_existed = _TrialComponent._load_or_create(
            trial_component_name=trial_component_name,
            display_name=run_display_name,
            tags=Run._append_run_tc_label_to_tags(tags),
            sagemaker_session=sagemaker_session,
        )
        if is_existed:
            logger.info(
                "The run (%s) under experiment (%s) already exists. Loading it. "
                "Note: sagemaker.experiments.load_run is recommended to use when "
                "the desired run already exists.",
                self.run_name,
                self.experiment_name,
            )
        self._trial.add_trial_component(self._trial_component)

        self._artifact_uploader = _ArtifactUploader(
            trial_component_name=self._trial_component.trial_component_name,
            sagemaker_session=sagemaker_session,
        )
        self._lineage_artifact_tracker = _LineageArtifactTracker(
            trial_component_arn=self._trial_component.trial_component_arn,
            sagemaker_session=sagemaker_session,
        )
        self._metrics_manager = _MetricsManager(
            trial_component_name=self._trial_component.trial_component_name,
            sagemaker_session=sagemaker_session,
        )
        self._inside_init_context = False
        self._inside_load_context = False
        self._in_load = False

    @property
    def experiment_config(self) -> dict:
        """Get experiment config from run attributes."""
        return {
            EXPERIMENT_NAME: self.experiment_name,
            TRIAL_NAME: self.run_group_name,
            RUN_NAME: self._trial_component.trial_component_name,
        }

    @validate_invoked_inside_run_context
    def log_parameter(self, name: str, value: Union[str, int, float]):
        """Record a single parameter value for this run.

        Overwrites any previous value recorded for the specified parameter name.

        Args:
            name (str): The name of the parameter.
            value (str or int or float): The value of the parameter.
        """
        if self._is_input_valid("parameter", name, value):
            self._trial_component.parameters[name] = value

    @validate_invoked_inside_run_context
    def log_parameters(self, parameters: Dict[str, Union[str, int, float]]):
        """Record a collection of parameter values for this run.

        Args:
            parameters (dict[str, str or int or float]): The parameters to record.
        """
        filtered_parameters = {
            key: value
            for (key, value) in parameters.items()
            if self._is_input_valid("parameter", key, value)
        }
        self._trial_component.parameters.update(filtered_parameters)

    @validate_invoked_inside_run_context
    def log_metric(
        self,
        name: str,
        value: float,
        timestamp: Optional[datetime.datetime] = None,
        step: Optional[int] = None,
    ):
        """Record a custom scalar metric value for this run.

        Note:
             This method is for manual custom metrics, for automatic metrics see the
             ``enable_sagemaker_metrics`` parameter on the ``estimator`` class.

        Args:
            name (str): The name of the metric.
            value (float): The value of the metric.
            timestamp (datetime.datetime): The timestamp of the metric.
                If not specified, the current UTC time will be used.
            step (int): The integer iteration number of the metric value (default: None).
        """
        if self._is_input_valid("metric", name, value):
            self._metrics_manager.log_metric(
                metric_name=name, value=value, timestamp=timestamp, step=step
            )

    @validate_invoked_inside_run_context
    def log_precision_recall(
        self,
        y_true: Union[list, array],
        predicted_probabilities: Union[list, array],
        positive_label: Optional[Union[str, int]] = None,
        title: Optional[str] = None,
        is_output: bool = True,
        no_skill: Optional[int] = None,
    ):
        """Create and log a precision recall graph artifact for Studio UI to render.

        The artifact is stored in S3 and represented as a lineage artifact
        with an association with the run.

        You can view the artifact in the UI.
        If your job is created by a pipeline execution you can view the artifact
        by selecting the corresponding step in the pipelines UI.
        See also `SageMaker Pipelines <https://aws.amazon.com/sagemaker/pipelines/>`_

        This method requires sklearn library.

        Args:
            y_true (list or array): True labels. If labels are not binary
                then positive_label should be given.
            predicted_probabilities (list or array): Estimated/predicted probabilities.
            positive_label (str or int): Label of the positive class (default: None).
            title (str): Title of the graph (default: None).
            is_output (bool): Determines direction of association to the
                run. Defaults to True (output artifact).
                If set to False then represented as input association.
            no_skill (int): The precision threshold under which the classifier cannot discriminate
                between the classes and would predict a random class or a constant class in
                all cases (default: None).
        """

        verify_length_of_true_and_predicted(
            true_labels=y_true,
            predicted_attrs=predicted_probabilities,
            predicted_attrs_name="predicted probabilities",
        )

        get_module("sklearn")
        from sklearn.metrics import precision_recall_curve, average_precision_score

        kwargs = {}
        if positive_label is not None:
            kwargs["pos_label"] = positive_label

        precision, recall, _ = precision_recall_curve(y_true, predicted_probabilities, **kwargs)

        kwargs["average"] = "micro"
        ap = average_precision_score(y_true, predicted_probabilities, **kwargs)

        data = {
            "type": "PrecisionRecallCurve",
            "version": 0,
            "title": title,
            "precision": precision.tolist(),
            "recall": recall.tolist(),
            "averagePrecisionScore": ap,
            "noSkill": no_skill,
        }
        self._log_graph_artifact(
            artifact_name=title, data=data, graph_type="PrecisionRecallCurve", is_output=is_output
        )

    @validate_invoked_inside_run_context
    def log_roc_curve(
        self,
        y_true: Union[list, array],
        y_score: Union[list, array],
        title: Optional[str] = None,
        is_output: bool = True,
    ):
        """Create and log a receiver operating characteristic (ROC curve) artifact.

        The artifact is stored in S3 and represented as a lineage artifact
        with an association with the run.

        You can view the artifact in the UI.
        If your job is created by a pipeline execution you can view the artifact
        by selecting the corresponding step in the pipelines UI.
        See also `SageMaker Pipelines <https://aws.amazon.com/sagemaker/pipelines/>`_

        This method requires sklearn library.

        Args:
            y_true (list or array): True labels. If labels are not binary
                then positive_label should be given.
            y_score (list or array): Estimated/predicted probabilities.
            title (str): Title of the graph (default: None).
            is_output (bool): Determines direction of association to the
                run. Defaults to True (output artifact).
                If set to False then represented as input association.
        """
        verify_length_of_true_and_predicted(
            true_labels=y_true, predicted_attrs=y_score, predicted_attrs_name="predicted scores"
        )

        get_module("sklearn")
        from sklearn.metrics import roc_curve, auc

        fpr, tpr, _ = roc_curve(y_true, y_score)

        auc = auc(fpr, tpr)

        data = {
            "type": "ROCCurve",
            "version": 0,
            "title": title,
            "falsePositiveRate": fpr.tolist(),
            "truePositiveRate": tpr.tolist(),
            "areaUnderCurve": auc,
        }
        self._log_graph_artifact(
            artifact_name=title, data=data, graph_type="ROCCurve", is_output=is_output
        )

    @validate_invoked_inside_run_context
    def log_confusion_matrix(
        self,
        y_true: Union[list, array],
        y_pred: Union[list, array],
        title: Optional[str] = None,
        is_output: bool = True,
    ):
        """Create and log a confusion matrix artifact.

        The artifact is stored in S3 and represented as a lineage artifact
        with an association with the run.

        You can view the artifact in the UI.
        If your job is created by a pipeline execution you can view the
        artifact by selecting the corresponding step in the pipelines UI.
        See also `SageMaker Pipelines <https://aws.amazon.com/sagemaker/pipelines/>`_
        This method requires sklearn library.

        Args:
            y_true (list or array): True labels. If labels are not binary
                then positive_label should be given.
            y_pred (list or array): Predicted labels.
            title (str): Title of the graph (default: None).
            is_output (bool): Determines direction of association to the
                run. Defaults to True (output artifact).
                If set to False then represented as input association.
        """
        verify_length_of_true_and_predicted(
            true_labels=y_true, predicted_attrs=y_pred, predicted_attrs_name="predicted labels"
        )

        get_module("sklearn")
        from sklearn.metrics import confusion_matrix

        matrix = confusion_matrix(y_true, y_pred)

        data = {
            "type": "ConfusionMatrix",
            "version": 0,
            "title": title,
            "confusionMatrix": matrix.tolist(),
        }
        self._log_graph_artifact(
            artifact_name=title, data=data, graph_type="ConfusionMatrix", is_output=is_output
        )

    @validate_invoked_inside_run_context
    def log_artifact(
        self, name: str, value: str, media_type: Optional[str] = None, is_output: bool = True
    ):
        """Record a single artifact for this run.

        Overwrites any previous value recorded for the specified name.

        Args:
            name (str): The name of the artifact.
            value (str): The value.
            media_type (str): The MediaType (MIME type) of the value (default: None).
            is_output (bool): Determines direction of association to the
                run. Defaults to True (output artifact).
                If set to False then represented as input association.
        """
        self._verify_trial_component_artifacts_length(is_output=is_output)
        if is_output:
            self._trial_component.output_artifacts[name] = TrialComponentArtifact(
                value, media_type=media_type
            )
        else:
            self._trial_component.input_artifacts[name] = TrialComponentArtifact(
                value, media_type=media_type
            )

    @validate_invoked_inside_run_context
    def log_file(
        self,
        file_path: str,
        name: Optional[str] = None,
        media_type: Optional[str] = None,
        is_output: bool = True,
    ):
        """Upload a file to s3 and store it as an input/output artifact in this run.

        Args:
            file_path (str): The path of the local file to upload.
            name (str): The name of the artifact (default: None).
            media_type (str): The MediaType (MIME type) of the file.
                If not specified, this library will attempt to infer the media type
                from the file extension of ``file_path``.
            is_output (bool): Determines direction of association to the
                run. Defaults to True (output artifact).
                If set to False then represented as input association.
        """
        self._verify_trial_component_artifacts_length(is_output)
        media_type = media_type or guess_media_type(file_path)
        name = name or resolve_artifact_name(file_path)
        s3_uri, _ = self._artifact_uploader.upload_artifact(file_path)
        if is_output:
            self._trial_component.output_artifacts[name] = TrialComponentArtifact(
                value=s3_uri, media_type=media_type
            )
        else:
            self._trial_component.input_artifacts[name] = TrialComponentArtifact(
                value=s3_uri, media_type=media_type
            )

    def close(self):
        """Persist any data saved locally."""
        try:
            # Update the trial component with additions from the Run object
            self._trial_component.save()
            # Create Lineage entities for the artifacts
            self._lineage_artifact_tracker.save()
        finally:
            if self._metrics_manager:
                self._metrics_manager.close()

    @staticmethod
    def _generate_trial_name(base_name) -> str:
        """Generate the reserved trial name based on experiment name

        Args:
            base_name (str): The ``experiment_name`` of this ``Run`` object.
        """
        available_length = MAX_NAME_LEN_IN_BACKEND - len(TRIAL_NAME_TEMPLATE)
        return TRIAL_NAME_TEMPLATE.format(base_name[:available_length])

    @staticmethod
    def _is_input_valid(input_type, field_name, field_value) -> bool:
        """Check if the input is valid or not

        Args:
            input_type (str): The type of the input, one of ``parameter``, ``metric``.
            field_name (str): The name of the field to be checked.
            field_value (str or int or float): The value of the field to be checked.
        """
        if isinstance(field_value, Number) and (isnan(field_value) or isinf(field_value)):
            logger.warning(
                "Failed to log %s %s. Received invalid value: %s.",
                input_type,
                field_name,
                field_value,
            )
            return False
        return True

    def _log_graph_artifact(self, data, graph_type, is_output, artifact_name=None):
        """Log an artifact.

        Logs an artifact by uploading data to S3, creating an artifact, and associating that
        artifact with the run trial component.

        Args:
            data (dict): Artifacts data that will be saved to S3.
            graph_type (str):  The type of the artifact.
            is_output (bool): Determines direction of association to the
                trial component. Defaults to True (output artifact).
                If set to False then represented as input association.
            artifact_name (str): Name of the artifact (default: None).
        """
        # generate an artifact name
        if not artifact_name:
            unique_name_from_base(graph_type)

        # create a json file in S3
        s3_uri, etag = self._artifact_uploader.upload_object_artifact(
            artifact_name, data, file_extension="json"
        )

        # create an artifact and association for the table
        if is_output:
            self._lineage_artifact_tracker.add_output_artifact(
                name=artifact_name, source_uri=s3_uri, etag=etag, artifact_type=graph_type
            )
        else:
            self._lineage_artifact_tracker.add_input_artifact(
                name=artifact_name, source_uri=s3_uri, etag=etag, artifact_type=graph_type
            )

    def _verify_trial_component_artifacts_length(self, is_output):
        """Verify the length of trial component artifacts

        Args:
            is_output (bool): Determines direction of association to the
                trial component.

        Raises:
            ValueError: If the length of trial component artifacts exceeds the limit.
        """
        err_msg_template = "Cannot add more than {} {}_artifacts under run"
        if is_output:
            if len(self._trial_component.output_artifacts) >= MAX_RUN_TC_ARTIFACTS_LEN:
                raise ValueError(err_msg_template.format(MAX_RUN_TC_ARTIFACTS_LEN, "output"))
        else:
            if len(self._trial_component.input_artifacts) >= MAX_RUN_TC_ARTIFACTS_LEN:
                raise ValueError(err_msg_template.format(MAX_RUN_TC_ARTIFACTS_LEN, "input"))

    @staticmethod
    def _generate_trial_component_name(run_name: str, experiment_name: str) -> str:
        """Generate the TrialComponentName based on run_name and experiment_name

        Args:
            run_name (str): The run_name supplied by the user.
            experiment_name (str): The experiment_name supplied by the user,
                which is prepended to the run_name to generate the TrialComponentName.

        Returns:
            str: The TrialComponentName used to create a trial component
                which is unique in an account.

        Raises:
            ValueError: If either the run_name or the experiment_name exceeds
                the length limit.
        """
        buffer = 1  # leave length buffers for delimiters
        max_len = int(MAX_NAME_LEN_IN_BACKEND / 2) - buffer
        err_msg_template = "The {} (length: {}) must have length less than or equal to {}"
        if len(run_name) > max_len:
            raise ValueError(err_msg_template.format("run_name", len(run_name), max_len))
        if len(experiment_name) > max_len:
            raise ValueError(
                err_msg_template.format("experiment_name", len(experiment_name), max_len)
            )
        trial_component_name = "{}{}{}".format(experiment_name, DELIMITER, run_name)
        # due to mixed-case concerns on the backend
        trial_component_name = trial_component_name.lower()
        return trial_component_name

    @staticmethod
    def _extract_run_name_from_tc_name(trial_component_name: str, experiment_name: str) -> str:
        """Extract the user supplied run name from a trial component name.

        Args:
            trial_component_name (str): The name of a run trial component.
            experiment_name (str): The experiment_name supplied by the user,
                which was prepended to the run_name to generate the trial_component_name.

        Returns:
            str: The name of the Run object supplied by a user.
        """
        return trial_component_name.replace("{}{}".format(experiment_name, DELIMITER), "", 1)

    @staticmethod
    def _append_run_tc_label_to_tags(tags: Optional[List[Dict[str, str]]] = None) -> list:
        """Append the run trial component label to tags used to create a trial component.

        Args:
            tags (List[Dict[str, str]]): The tags supplied by users to initialize a Run object.

        Returns:
            list: The updated tags with the appended run trial component label.
        """
        if not tags:
            tags = []
        tags.append(RUN_TC_TAG)
        return tags

    def __enter__(self):
        """Updates the start time of the run.

        Returns:
            object: self.
        """
        nested_with_err_msg_template = (
            "It is not allowed to use nested 'with' statements on the {}."
        )
        if self._in_load:
            if self._inside_load_context:
                raise RuntimeError(nested_with_err_msg_template.format("load_run"))
            self._inside_load_context = True
        else:
            if _RunContext.get_current_run():
                raise RuntimeError(nested_with_err_msg_template.format("Run"))
            self._inside_init_context = True
            _RunContext.add_run_object(self)

        if not self._trial_component.start_time:
            start_time = datetime.datetime.now(dateutil.tz.tzlocal())
            self._trial_component.start_time = start_time
        self._trial_component.status = _api_types.TrialComponentStatus(
            primary_status=_TrialComponentStatusType.InProgress.value,
            message="Within a run context",
        )
        # Save the start_time and status changes to backend
        self._trial_component.save()
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Updates the end time of the run.

        Args:
            exc_type (str): The exception type.
            exc_value (str): The exception value.
            exc_traceback (str): The stack trace of the exception.
        """
        if self._in_load:
            self._inside_load_context = False
            self._in_load = False
        else:
            self._inside_init_context = False
            _RunContext.drop_current_run()

        end_time = datetime.datetime.now(dateutil.tz.tzlocal())
        self._trial_component.end_time = end_time
        if exc_value:
            self._trial_component.status = _api_types.TrialComponentStatus(
                primary_status=_TrialComponentStatusType.Failed.value, message=str(exc_value)
            )
        else:
            self._trial_component.status = _api_types.TrialComponentStatus(
                primary_status=_TrialComponentStatusType.Completed.value
            )

        self.close()


def load_run(
    run_name: Optional[str] = None,
    experiment_name: Optional[str] = None,
    sagemaker_session: Optional["Session"] = None,
) -> Run:
    """Load an existing run.

    In order to reuse an existing run to log extra data, ``load_run`` is recommended.
    It can be used in several ways:

    1. Use ``load_run`` by explicitly passing in ``run_name`` and ``experiment_name``.

    If ``run_name`` and ``experiment_name`` are passed in, they are honored over
    the default experiment config in the job environment or the run context
    (i.e. within the ``with`` block).

    Note:
        Both ``run_name`` and ``experiment_name`` should be supplied to make this usage work.
        Otherwise, you may get a ``ValueError``.

    .. code:: python

        with load_run(experiment_name="my-exp", run_name="my-run") as run:
            run.log_metric(...)
            ...

    2. Use the ``load_run`` in a job script without supplying ``run_name`` and ``experiment_name``.

    In this case, the default experiment config (specified when creating the job) is fetched
    from the job environment to load the run.

    .. code:: python

        # In a job script
        with load_run() as run:
            run.log_metric(...)
            ...

    3. Use the ``load_run`` in a notebook within a run context (i.e. the ``with`` block)
    but without supplying ``run_name`` and ``experiment_name``.

    Every time we call ``with Run(...) as run1:``, the initialized ``run1`` is tracked
    in the run context. Then when we call ``load_run()`` under this with statement, the ``run1``
    in the context is loaded by default.

    .. code:: python

        # In a notebook
        with Run(experiment_name="my-exp", run_name="my-run", ...) as run1:
            run1.log_parameter(...)

            with load_run() as run2: # run2 is the same object as run1
                run2.log_metric(...)
                ...

    Args:
        run_name (str): The name of the run to be loaded (default: None).
            If it is None, the ``RunName`` in the ``ExperimentConfig`` of the job will be
            fetched to load the run.
        experiment_name (str): The name of the Experiment that the to be loaded run
            is associated with (default: None).
            Note: the experiment_name must be supplied along with a valid run_name.
            Otherwise, it will be ignored.
        sagemaker_session (sagemaker.session.Session): Session object which
            manages interactions with Amazon SageMaker APIs and any other
            AWS services needed. If not specified, one is created using the
            default AWS configuration chain.

    Returns:
        Run: The loaded Run object.
    """
    sagemaker_session = sagemaker_session or _utils.default_session()
    environment = _RunEnvironment.load()

    verify_load_input_names(run_name=run_name, experiment_name=experiment_name)

    if run_name or environment:
        if run_name:
            logger.warning(
                "run_name is explicitly supplied in load_run, "
                "which will be prioritized to load the Run object. "
                "In other words, the run name in the experiment config, fetched from the "
                "job environment or the current run context, will be ignored."
            )
        else:
            exp_config = get_tc_and_exp_config_from_job_env(
                environment=environment, sagemaker_session=sagemaker_session
            )
            run_name = Run._extract_run_name_from_tc_name(
                trial_component_name=exp_config[RUN_NAME],
                experiment_name=exp_config[EXPERIMENT_NAME],
            )
            experiment_name = exp_config[EXPERIMENT_NAME]

        run_instance = Run(
            experiment_name=experiment_name,
            run_name=run_name,
            sagemaker_session=sagemaker_session,
        )
    elif _RunContext.get_current_run():
        run_instance = _RunContext.get_current_run()
    else:
        raise RuntimeError(
            "Failed to load a Run object. "
            "Please make sure a Run object has been initialized already."
        )

    run_instance._in_load = True
    return run_instance


def list_runs(
    experiment_name: str,
    created_before: Optional[datetime.datetime] = None,
    created_after: Optional[datetime.datetime] = None,
    sagemaker_session: Optional["Session"] = None,
    max_results: Optional[int] = None,
    next_token: Optional[str] = None,
    sort_by: SortByType = SortByType.CREATION_TIME,
    sort_order: SortOrderType = SortOrderType.DESCENDING,
) -> list:
    """Return a list of ``Run`` objects matching the given criteria.

    Args:
        experiment_name (str): Only Run objects related to the specified experiment
            are returned.
        created_before (datetime.datetime): Return Run objects created before this instant
            (default: None).
        created_after (datetime.datetime): Return Run objects created after this instant
            (default: None).
        sagemaker_session (sagemaker.session.Session): Session object which
            manages interactions with Amazon SageMaker APIs and any other
            AWS services needed. If not specified, one is created using the
            default AWS configuration chain.
        max_results (int): Maximum number of Run objects to retrieve (default: None).
        next_token (str): Token for next page of results (default: None).
        sort_by (SortByType): The property to sort results by. One of NAME, CREATION_TIME
            (default: CREATION_TIME).
        sort_order (SortOrderType): One of ASCENDING, or DESCENDING (default: DESCENDING).

    Returns:
        list: A list of ``Run`` objects.
    """
    tc_summaries = _TrialComponent.list(
        experiment_name=experiment_name,
        created_before=created_before,
        created_after=created_after,
        sort_by=sort_by.value,
        sort_order=sort_order.value,
        sagemaker_session=sagemaker_session,
        max_results=max_results,
        next_token=next_token,
    )
    run_list = []
    for tc_summary in tc_summaries:
        if not is_run_trial_component(
            trial_component_name=tc_summary.trial_component_name,
            sagemaker_session=sagemaker_session,
        ):
            continue
        run_instance = Run(
            experiment_name=experiment_name,
            run_name=Run._extract_run_name_from_tc_name(
                trial_component_name=tc_summary.trial_component_name,
                experiment_name=experiment_name,
            ),
            sagemaker_session=sagemaker_session,
        )
        run_list.append(run_instance)
    return run_list



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Scrapper utilities to support repacking of models."""
from __future__ import absolute_import

import logging
import os
import shutil
import tarfile
import tempfile
from typing import List, Union, Optional, TYPE_CHECKING
from sagemaker import image_uris
from sagemaker.inputs import TrainingInput
from sagemaker.estimator import EstimatorBase
from sagemaker.sklearn.estimator import SKLearn
from sagemaker.workflow.entities import RequestType
from sagemaker.workflow.properties import Properties
from sagemaker.session import get_create_model_package_request, get_model_package_args
from sagemaker.workflow.steps import (
    StepTypeEnum,
    TrainingStep,
    Step,
    ConfigurableRetryStep,
)
from sagemaker.utils import _save_model, download_file_from_url
from sagemaker.workflow.retry import RetryPolicy

if TYPE_CHECKING:
    from sagemaker.workflow.step_collections import StepCollection

logger = logging.getLogger(__name__)

FRAMEWORK_VERSION = "0.23-1"
INSTANCE_TYPE = "ml.m5.large"
REPACK_SCRIPT = "_repack_model.py"
REPACK_SCRIPT_LAUNCHER = "_repack_script_launcher.sh"
LAUNCH_REPACK_SCRIPT_CMD = """
#!/bin/bash

var_dependencies="${SM_HP_DEPENDENCIES}"
var_inference_script="${SM_HP_INFERENCE_SCRIPT}"
var_model_archive="${SM_HP_MODEL_ARCHIVE}"
var_source_dir="${SM_HP_SOURCE_DIR}"

python _repack_model.py \
--dependencies "${var_dependencies}" \
--inference_script "${var_inference_script}" \
--model_archive "${var_model_archive}" \
--source_dir "${var_source_dir}"
"""


class _RepackModelStep(TrainingStep):
    """Repacks model artifacts with custom inference entry points.

    The SDK automatically adds this step to pipelines that have RegisterModelSteps with models
    that have a custom entry point.
    """

    def __init__(
        self,
        name: str,
        sagemaker_session,
        role,
        model_data: str,
        entry_point: str,
        display_name: str = None,
        description: str = None,
        source_dir: str = None,
        dependencies: List = None,
        depends_on: Optional[List[Union[str, Step, "StepCollection"]]] = None,
        retry_policies: List[RetryPolicy] = None,
        subnets=None,
        security_group_ids=None,
        **kwargs,
    ):
        """Base class initializer.

        Args:
            name (str): The name of the training step.
            sagemaker_session (sagemaker.session.Session): Session object which manages
                    interactions with Amazon SageMaker APIs and any other AWS services needed. If
                    not specified, the estimator creates one using the default
                    AWS configuration chain.
            role (str): An AWS IAM role (either name or full ARN). The Amazon
                    SageMaker training jobs and APIs that create Amazon SageMaker
                    endpoints use this role to access training data and model
                    artifacts. After the endpoint is created, the inference code
                    might use the IAM role, if it needs to access an AWS resource.
            model_data (str): The S3 location of a SageMaker model data `.tar.gz` file.
            entry_point (str): Path (absolute or relative) to the local Python
                    source file which should be executed as the entry point to
                    inference. If ``source_dir`` is specified, then ``entry_point``
                    must point to a file located at the root of ``source_dir``.
                    If 'git_config' is provided, 'entry_point' should be
                    a relative location to the Python source file in the Git repo.

                    Example:
                        With the following GitHub repo directory structure:

                        >>> |----- README.md
                        >>> |----- src
                        >>>         |----- train.py
                        >>>         |----- test.py

                        You can assign entry_point='src/train.py'.
            display_name (str): The display name of this `_RepackModelStep` step (default: None).
            description (str): The description of this `_RepackModelStep` (default: None).
            source_dir (str): A relative location to a directory with other training
                or model hosting source code dependencies aside from the entry point
                file in the Git repo (default: None). Structure within this
                directory are preserved when training on Amazon SageMaker.
            dependencies (list[str]): A list of paths to directories (absolute
                    or relative) with any additional libraries that will be exported
                    to the container (default: []). The library folders will be
                    copied to SageMaker in the same folder where the entrypoint is
                    copied. If 'git_config' is provided, 'dependencies' should be a
                    list of relative locations to directories with any additional
                    libraries needed in the Git repo.

                    .. admonition:: Example

                        The following call

                        >>> Estimator(entry_point='train.py',
                        ...           dependencies=['my/libs/common', 'virtual-env'])

                        results in the following inside the container:

                        >>> $ ls

                        >>> opt/ml/code
                        >>>     |------ train.py
                        >>>     |------ common
                        >>>     |------ virtual-env

                    This is not supported with "local code" in Local Mode.
            depends_on (List[Union[str, Step, StepCollection]]): The list of `Step`/`StepCollection`
                names or `Step` instances or `StepCollection` instances that the current `Step`
                depends on (default: None).
            retry_policies (List[RetryPolicy]): The list of retry policies for the current step
                (default: None).
            subnets (list[str]): List of subnet ids. If not specified, the re-packing
                    job will be created without VPC config (default: None).
            security_group_ids (list[str]): List of security group ids. If not
                specified, the re-packing job will be created without VPC config (default: None).
            **kwargs: additional arguments for the repacking job.
        """
        self._model_data = model_data
        self.sagemaker_session = sagemaker_session
        self.role = role
        self._entry_point = entry_point
        self._entry_point_basename = os.path.basename(self._entry_point)
        self._source_dir = source_dir
        self._dependencies = dependencies

        # convert dependencies array into space-delimited string
        dependencies_hyperparameter = None
        if self._dependencies:
            dependencies_hyperparameter = " ".join(self._dependencies)

        # the real estimator and inputs
        repacker = SKLearn(
            framework_version=FRAMEWORK_VERSION,
            instance_type=INSTANCE_TYPE,
            entry_point=REPACK_SCRIPT_LAUNCHER,
            source_dir=self._source_dir,
            dependencies=self._dependencies,
            sagemaker_session=self.sagemaker_session,
            role=self.role,
            hyperparameters={
                "inference_script": self._entry_point_basename,
                "model_archive": self._model_data,
                "dependencies": dependencies_hyperparameter,
                "source_dir": self._source_dir,
            },
            subnets=subnets,
            security_group_ids=security_group_ids,
            **kwargs,
        )
        repacker.disable_profiler = True
        inputs = TrainingInput(self._model_data)

        # super!
        super(_RepackModelStep, self).__init__(
            name=name,
            display_name=display_name,
            description=description,
            depends_on=depends_on,
            retry_policies=retry_policies,
            estimator=repacker,
            inputs=inputs,
        )

    def _prepare_for_repacking(self):
        """Prepares the source for the estimator."""
        if self._source_dir is None:
            self._establish_source_dir()

        self._inject_repack_script_and_launcher()

    def _establish_source_dir(self):
        """If the source_dir is None, creates it for the repacking job.

        It performs the following:
            1) creates a source directory
            2) copies the inference_entry_point inside it
            3) copies the repack_model.py inside it
            4) sets the source dir for the repacking estimator
        """
        self._source_dir = tempfile.mkdtemp()
        self.estimator.source_dir = self._source_dir

        shutil.copy2(self._entry_point, os.path.join(self._source_dir, self._entry_point_basename))
        self._entry_point = self._entry_point_basename

    def _inject_repack_script_and_launcher(self):
        """Injects the _repack_model.py script and _repack_script_launcher.sh

        into S3 or local source directory.

        Note: The bash file is needed because if not supplied, the SKLearn
        training job will auto install all dependencies listed in requirements.txt.
        However, this auto install behavior is not expected in _RepackModelStep,
        since it should just simply repack the model along with other supplied files,
        e.g. the requirements.txt.

        If the source_dir is an S3 path:
            1) downloads the source_dir tar.gz
            2) extracts it
            3) copies the _repack_model.py script into the extracted directory
            4) creates the _repack_script_launcher.sh in the extracted dir
            5) rezips the directory
            6) overwrites the S3 source_dir with the new tar.gz

        If the source_dir is a local path:
            1) copies the _repack_model.py script into the source dir
            2) creates the _repack_script_launcher.sh in the source dir
        """
        fname = os.path.join(os.path.dirname(__file__), REPACK_SCRIPT)
        if self._source_dir.lower().startswith("s3://"):
            with tempfile.TemporaryDirectory() as tmp:
                targz_contents_dir = os.path.join(tmp, "extracted")

                old_targz_path = os.path.join(tmp, "old.tar.gz")
                download_file_from_url(self._source_dir, old_targz_path, self.sagemaker_session)

                with tarfile.open(name=old_targz_path, mode="r:gz") as t:
                    t.extractall(path=targz_contents_dir)

                shutil.copy2(fname, os.path.join(targz_contents_dir, REPACK_SCRIPT))
                with open(
                    os.path.join(targz_contents_dir, REPACK_SCRIPT_LAUNCHER), "w"
                ) as launcher_file:
                    launcher_file.write(LAUNCH_REPACK_SCRIPT_CMD)

                new_targz_path = os.path.join(tmp, "new.tar.gz")
                with tarfile.open(new_targz_path, mode="w:gz") as t:
                    t.add(targz_contents_dir, arcname=os.path.sep)

                _save_model(self._source_dir, new_targz_path, self.sagemaker_session, kms_key=None)
        else:
            shutil.copy2(fname, os.path.join(self._source_dir, REPACK_SCRIPT))
            with open(os.path.join(self._source_dir, REPACK_SCRIPT_LAUNCHER), "w") as launcher_file:
                launcher_file.write(LAUNCH_REPACK_SCRIPT_CMD)

    @property
    def arguments(self) -> RequestType:
        """The arguments dict that are used to call `create_training_job`.

        This first prepares the source bundle for repackinglby placing artifacts
        in locations which the training container will make available to the
        repacking script and then gets the arguments for the training job.
        """
        self._prepare_for_repacking()
        return super(_RepackModelStep, self).arguments

    @property
    def properties(self):
        """A Properties object representing the DescribeTrainingJobResponse data model."""
        return self._properties


class _RegisterModelStep(ConfigurableRetryStep):
    """Register model step in workflow that creates a model package."""

    def __init__(
        self,
        name: str,
        step_args: Optional[dict] = None,
        content_types: Optional[list] = None,
        response_types: Optional[list] = None,
        inference_instances: Optional[list] = None,
        transform_instances: Optional[list] = None,
        estimator: EstimatorBase = None,
        model_data=None,
        model_package_group_name=None,
        model_metrics=None,
        metadata_properties=None,
        approval_status="PendingManualApproval",
        image_uri=None,
        compile_model_family=None,
        display_name: str = None,
        description=None,
        depends_on: Optional[List[Union[str, Step, "StepCollection"]]] = None,
        retry_policies: Optional[List[RetryPolicy]] = None,
        tags=None,
        container_def_list=None,
        drift_check_baselines=None,
        customer_metadata_properties=None,
        domain=None,
        sample_payload_url=None,
        task=None,
        **kwargs,
    ):
        """Constructor of a register model step.

        Args:
            name (str): The name of the training step.
            step_args (dict): The arguments for this `_RegisterModelStep` definition
                (default: None).
            content_types (list): The supported MIME types for the input data (default: None).
            response_types (list): The supported MIME types for the output data (default: None).
            inference_instances (list): A list of the instance types that are used to
                generate inferences in real-time (default: None).
            transform_instances (list): A list of the instance types on which a
                transformation job can be run or on which an endpoint
                can be deployed (default: None).
            estimator (EstimatorBase): A `sagemaker.estimator.EstimatorBase` instance
                (default: None).
            model_data: the S3 URI to the model data from training (default: None).
            model_package_group_name (str): Model Package Group name, exclusive to
                `model_package_name`, using `model_package_group_name`
                makes the Model Package versioned (default: None).
            model_metrics (ModelMetrics): ModelMetrics object (default: None).
            metadata_properties (MetadataProperties): MetadataProperties object (default: None).
            approval_status (str): Model Approval Status, values can be "Approved",
                "Rejected", or "PendingManualApproval" (default: "PendingManualApproval").
            image_uri (str): The container image uri for Model Package, if not specified,
                Estimator's training container image will be used (default: None).
            compile_model_family (str): Instance family for compiled model,
                if specified, a compiled model will be used (default: None).
            display_name (str): The display name of this `_RegisterModelStep` step (default: None).
            description (str): Model Package description (default: None).
            depends_on (List[Union[str, Step, StepCollection]]): The list of `Step`/`StepCollection`
                names or `Step` instances or `StepCollection` instances that the current `Step`
                depends on (default: None).
            retry_policies (List[RetryPolicy]): The list of retry policies for the current step
                (default: None).
            tags (List[dict[str, str]]): A list of dictionaries containing key-value pairs used to
                configure the create model package request (default: None).
            container_def_list (list): A list of container definitions (default: None).
            drift_check_baselines (DriftCheckBaselines): DriftCheckBaselines object (default: None).
            customer_metadata_properties (dict[str, str]): A dictionary of key-value paired
                metadata properties (default: None).
            domain (str): Domain values can be "COMPUTER_VISION", "NATURAL_LANGUAGE_PROCESSING",
                "MACHINE_LEARNING" (default: None).
            sample_payload_url (str): The S3 path where the sample payload is stored
                (default: None).
            task (str): Task values which are supported by Inference Recommender are "FILL_MASK",
                "IMAGE_CLASSIFICATION", "OBJECT_DETECTION", "TEXT_GENERATION", "IMAGE_SEGMENTATION",
                "CLASSIFICATION", "REGRESSION", "OTHER" (default: None).
            **kwargs: additional arguments to `create_model`.
        """
        super(_RegisterModelStep, self).__init__(
            name, StepTypeEnum.REGISTER_MODEL, display_name, description, depends_on, retry_policies
        )
        deprecated_args_missing = content_types is None or response_types is None
        if not (step_args is None) ^ deprecated_args_missing:
            raise ValueError(
                "step_args and the set of (content_types, response_types, "
                ") are mutually exclusive. "
                "Either of them should be provided."
            )

        self.step_args = step_args
        self.estimator = estimator
        self.model_data = model_data
        self.content_types = content_types
        self.response_types = response_types
        self.inference_instances = inference_instances
        self.transform_instances = transform_instances
        self.model_package_group_name = model_package_group_name
        self.tags = tags
        self.model_metrics = model_metrics
        self.drift_check_baselines = drift_check_baselines
        self.customer_metadata_properties = customer_metadata_properties
        self.domain = domain
        self.sample_payload_url = sample_payload_url
        self.task = task
        self.metadata_properties = metadata_properties
        self.approval_status = approval_status
        self.image_uri = image_uri
        self.compile_model_family = compile_model_family
        self.description = description
        self.tags = tags
        self.kwargs = kwargs
        self.container_def_list = container_def_list

        self._properties = Properties(step_name=name, shape_name="DescribeModelPackageOutput")

    @property
    def arguments(self) -> RequestType:
        """The arguments dict that are used to call `create_model_package`."""
        model_name = self.name

        if self.step_args:
            request_dict = self.step_args
        else:
            if self.container_def_list is None:
                if self.compile_model_family:
                    model = self.estimator._compiled_models[self.compile_model_family]
                    self.model_data = model.model_data
                else:
                    # create_model wants the estimator to have a model_data attribute...
                    self.estimator._current_job_name = model_name

                    # placeholder. replaced with model_data later
                    output_path = self.estimator.output_path
                    self.estimator.output_path = "/tmp"

                    # create the model, but custom funky framework stuff going on in some places
                    if self.image_uri:
                        model = self.estimator.create_model(image_uri=self.image_uri, **self.kwargs)
                    else:
                        model = self.estimator.create_model(**self.kwargs)
                        self.image_uri = model.image_uri

                    if self.model_data is None:
                        self.model_data = model.model_data

                    # reset placeholder
                    self.estimator.output_path = output_path

                    # yeah, there is some framework stuff going on that we need to pull in here
                    if self.image_uri is None:
                        region_name = self.estimator.sagemaker_session.boto_session.region_name
                        self.image_uri = image_uris.retrieve(
                            model._framework_name,
                            region_name,
                            version=model.framework_version,
                            py_version=model.py_version if hasattr(model, "py_version") else None,
                            instance_type=self.kwargs.get(
                                "instance_type", self.estimator.instance_type
                            ),
                            accelerator_type=self.kwargs.get("accelerator_type"),
                            image_scope="inference",
                        )
                        model.name = model_name
                        model.model_data = self.model_data

            model_package_args = get_model_package_args(
                content_types=self.content_types,
                response_types=self.response_types,
                inference_instances=self.inference_instances,
                transform_instances=self.transform_instances,
                model_package_group_name=self.model_package_group_name,
                model_data=self.model_data,
                image_uri=self.image_uri,
                model_metrics=self.model_metrics,
                drift_check_baselines=self.drift_check_baselines,
                metadata_properties=self.metadata_properties,
                approval_status=self.approval_status,
                description=self.description,
                tags=self.tags,
                container_def_list=self.container_def_list,
                customer_metadata_properties=self.customer_metadata_properties,
                domain=self.domain,
                sample_payload_url=self.sample_payload_url,
                task=self.task,
            )

            request_dict = get_create_model_package_request(**model_package_args)
        # these are not available in the workflow service and will cause rejection
        warn_msg_template = (
            "Popping out '%s' from the pipeline definition "
            "since it will be overridden in pipeline execution time."
        )
        if "CertifyForMarketplace" in request_dict:
            request_dict.pop("CertifyForMarketplace")
            logger.warning(warn_msg_template, "CertifyForMarketplace")
        if "Description" in request_dict:
            request_dict.pop("Description")
            logger.warning(warn_msg_template, "Description")
        if "ModelPackageName" in request_dict:
            request_dict.pop("ModelPackageName")
            logger.warning(warn_msg_template, "ModelPackageName")

        return request_dict

    @property
    def properties(self):
        """A Properties object representing the DescribeModelPackageOutput data model."""
        return self._properties



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""This module contains functions for obtaining JumpStart ECR and S3 URIs."""
from __future__ import absolute_import
import os
from typing import Dict, Optional
from sagemaker import image_uris
from sagemaker.jumpstart.constants import (
    ENV_VARIABLE_JUMPSTART_MODEL_ARTIFACT_BUCKET_OVERRIDE,
    ENV_VARIABLE_JUMPSTART_SCRIPT_ARTIFACT_BUCKET_OVERRIDE,
    JUMPSTART_DEFAULT_REGION_NAME,
)
from sagemaker.jumpstart.enums import (
    JumpStartScriptScope,
    ModelFramework,
    VariableScope,
)
from sagemaker.jumpstart.utils import (
    get_jumpstart_content_bucket,
    verify_model_region_and_return_specs,
)
from sagemaker.jumpstart import accessors as jumpstart_accessors


def _retrieve_image_uri(
    model_id: str,
    model_version: str,
    image_scope: str,
    framework: Optional[str],
    region: Optional[str],
    version: Optional[str],
    py_version: Optional[str],
    instance_type: Optional[str],
    accelerator_type: Optional[str],
    container_version: Optional[str],
    distribution: Optional[str],
    base_framework_version: Optional[str],
    training_compiler_config: Optional[str],
    tolerate_vulnerable_model: bool,
    tolerate_deprecated_model: bool,
):
    """Retrieves the container image URI for JumpStart models.

    Only `model_id`, `model_version`, and `image_scope` are required;
    the rest of the fields are auto-populated.


    Args:
        model_id (str): JumpStart model ID for which to retrieve image URI.
        model_version (str): Version of the JumpStart model for which to retrieve
            the image URI.
        image_scope (str): The image type, i.e. what it is used for.
            Valid values: "training", "inference", "eia". If ``accelerator_type`` is set,
            ``image_scope`` is ignored.
        framework (str): The name of the framework or algorithm.
        region (str): The AWS region.
        version (str): The framework or algorithm version. This is required if there is
            more than one supported version for the given framework or algorithm.
        py_version (str): The Python version. This is required if there is
            more than one supported Python version for the given framework version.
        instance_type (str): The SageMaker instance type. For supported types, see
            https://aws.amazon.com/sagemaker/pricing/instance-types. This is required if
            there are different images for different processor types.
        accelerator_type (str): Elastic Inference accelerator type. For more, see
            https://docs.aws.amazon.com/sagemaker/latest/dg/ei.html.
        container_version (str): the version of docker image.
            Ideally the value of parameter should be created inside the framework.
            For custom use, see the list of supported container versions:
            https://github.com/aws/deep-learning-containers/blob/master/available_images.md.
        distribution (dict): A dictionary with information on how to run distributed training
        training_compiler_config (:class:`~sagemaker.training_compiler.TrainingCompilerConfig`):
            A configuration class for the SageMaker Training Compiler.
        tolerate_vulnerable_model (bool): True if vulnerable versions of model
            specifications should be tolerated (exception not raised). If False, raises an
            exception if the script used by this version of the model has dependencies with known
            security vulnerabilities.
        tolerate_deprecated_model (bool): True if deprecated versions of model
            specifications should be tolerated (exception not raised). If False, raises
            an exception if the version of the model is deprecated.

    Returns:
        str: the ECR URI for the corresponding SageMaker Docker image.

    Raises:
        ValueError: If the combination of arguments specified is not supported.
        VulnerableJumpStartModelError: If any of the dependencies required by the script have
            known security vulnerabilities.
        DeprecatedJumpStartModelError: If the version of the model is deprecated.
    """
    if region is None:
        region = JUMPSTART_DEFAULT_REGION_NAME

    model_specs = verify_model_region_and_return_specs(
        model_id=model_id,
        version=model_version,
        scope=image_scope,
        region=region,
        tolerate_vulnerable_model=tolerate_vulnerable_model,
        tolerate_deprecated_model=tolerate_deprecated_model,
    )

    if image_scope == JumpStartScriptScope.INFERENCE:
        ecr_specs = model_specs.hosting_ecr_specs
    elif image_scope == JumpStartScriptScope.TRAINING:
        ecr_specs = model_specs.training_ecr_specs

    if framework is not None and framework != ecr_specs.framework:
        raise ValueError(
            f"Incorrect container framework '{framework}' for JumpStart model ID '{model_id}' "
            f"and version '{model_version}'."
        )

    if version is not None and version != ecr_specs.framework_version:
        raise ValueError(
            f"Incorrect container framework version '{version}' for JumpStart model ID "
            f"'{model_id}' and version '{model_version}'."
        )

    if py_version is not None and py_version != ecr_specs.py_version:
        raise ValueError(
            f"Incorrect python version '{py_version}' for JumpStart model ID '{model_id}' "
            f"and version '{model_version}'."
        )

    base_framework_version_override: Optional[str] = None
    version_override: Optional[str] = None
    if ecr_specs.framework == ModelFramework.HUGGINGFACE:
        base_framework_version_override = ecr_specs.framework_version
        version_override = ecr_specs.huggingface_transformers_version

    if image_scope == JumpStartScriptScope.TRAINING:
        return image_uris.get_training_image_uri(
            region=region,
            framework=ecr_specs.framework,
            framework_version=version_override or ecr_specs.framework_version,
            py_version=ecr_specs.py_version,
            image_uri=None,
            distribution=None,
            compiler_config=None,
            tensorflow_version=None,
            pytorch_version=base_framework_version_override or base_framework_version,
            instance_type=instance_type,
        )
    if base_framework_version_override is not None:
        base_framework_version_override = f"pytorch{base_framework_version_override}"

    return image_uris.retrieve(
        framework=ecr_specs.framework,
        region=region,
        version=version_override or ecr_specs.framework_version,
        py_version=ecr_specs.py_version,
        instance_type=instance_type,
        accelerator_type=accelerator_type,
        image_scope=image_scope,
        container_version=container_version,
        distribution=distribution,
        base_framework_version=base_framework_version_override or base_framework_version,
        training_compiler_config=training_compiler_config,
    )


def _retrieve_model_uri(
    model_id: str,
    model_version: str,
    model_scope: Optional[str],
    region: Optional[str],
    tolerate_vulnerable_model: bool,
    tolerate_deprecated_model: bool,
):
    """Retrieves the model artifact S3 URI for the model matching the given arguments.

    Optionally uses a bucket override specified by environment variable.

    Args:
        model_id (str): JumpStart model ID of the JumpStart model for which to retrieve
            the model artifact S3 URI.
        model_version (str): Version of the JumpStart model for which to retrieve the model
            artifact S3 URI.
        model_scope (str): The model type, i.e. what it is used for.
            Valid values: "training" and "inference".
        region (str): Region for which to retrieve model S3 URI.
        tolerate_vulnerable_model (bool): True if vulnerable versions of model
            specifications should be tolerated (exception not raised). If False, raises an
            exception if the script used by this version of the model has dependencies with known
            security vulnerabilities.
        tolerate_deprecated_model (bool): True if deprecated versions of model
            specifications should be tolerated (exception not raised). If False, raises
            an exception if the version of the model is deprecated.
    Returns:
        str: the model artifact S3 URI for the corresponding model.

    Raises:
        ValueError: If the combination of arguments specified is not supported.
        VulnerableJumpStartModelError: If any of the dependencies required by the script have
            known security vulnerabilities.
        DeprecatedJumpStartModelError: If the version of the model is deprecated.
    """
    if region is None:
        region = JUMPSTART_DEFAULT_REGION_NAME

    model_specs = verify_model_region_and_return_specs(
        model_id=model_id,
        version=model_version,
        scope=model_scope,
        region=region,
        tolerate_vulnerable_model=tolerate_vulnerable_model,
        tolerate_deprecated_model=tolerate_deprecated_model,
    )

    if model_scope == JumpStartScriptScope.INFERENCE:
        model_artifact_key = model_specs.hosting_artifact_key
    elif model_scope == JumpStartScriptScope.TRAINING:
        model_artifact_key = model_specs.training_artifact_key

    bucket = os.environ.get(
        ENV_VARIABLE_JUMPSTART_MODEL_ARTIFACT_BUCKET_OVERRIDE
    ) or get_jumpstart_content_bucket(region)

    model_s3_uri = f"s3://{bucket}/{model_artifact_key}"

    return model_s3_uri


def _retrieve_script_uri(
    model_id: str,
    model_version: str,
    script_scope: Optional[str],
    region: Optional[str],
    tolerate_vulnerable_model: bool,
    tolerate_deprecated_model: bool,
):
    """Retrieves the script S3 URI associated with the model matching the given arguments.

    Optionally uses a bucket override specified by environment variable.

    Args:
        model_id (str): JumpStart model ID of the JumpStart model for which to
            retrieve the script S3 URI.
        model_version (str): Version of the JumpStart model for which to
            retrieve the model script S3 URI.
        script_scope (str): The script type, i.e. what it is used for.
            Valid values: "training" and "inference".
        region (str): Region for which to retrieve model script S3 URI.
        tolerate_vulnerable_model (bool): True if vulnerable versions of model
            specifications should be tolerated (exception not raised). If False, raises an
            exception if the script used by this version of the model has dependencies with known
            security vulnerabilities.
        tolerate_deprecated_model (bool): True if deprecated versions of model
            specifications should be tolerated (exception not raised). If False, raises
            an exception if the version of the model is deprecated.
    Returns:
        str: the model script URI for the corresponding model.

    Raises:
        ValueError: If the combination of arguments specified is not supported.
        VulnerableJumpStartModelError: If any of the dependencies required by the script have
            known security vulnerabilities.
        DeprecatedJumpStartModelError: If the version of the model is deprecated.
    """
    if region is None:
        region = JUMPSTART_DEFAULT_REGION_NAME

    model_specs = verify_model_region_and_return_specs(
        model_id=model_id,
        version=model_version,
        scope=script_scope,
        region=region,
        tolerate_vulnerable_model=tolerate_vulnerable_model,
        tolerate_deprecated_model=tolerate_deprecated_model,
    )

    if script_scope == JumpStartScriptScope.INFERENCE:
        model_script_key = model_specs.hosting_script_key
    elif script_scope == JumpStartScriptScope.TRAINING:
        model_script_key = model_specs.training_script_key

    bucket = os.environ.get(
        ENV_VARIABLE_JUMPSTART_SCRIPT_ARTIFACT_BUCKET_OVERRIDE
    ) or get_jumpstart_content_bucket(region)

    script_s3_uri = f"s3://{bucket}/{model_script_key}"

    return script_s3_uri


def _retrieve_default_hyperparameters(
    model_id: str,
    model_version: str,
    region: Optional[str],
    include_container_hyperparameters: bool = False,
):
    """Retrieves the training hyperparameters for the model matching the given arguments.

    Args:
        model_id (str): JumpStart model ID of the JumpStart model for which to
            retrieve the default hyperparameters.
        model_version (str): Version of the JumpStart model for which to retrieve the
            default hyperparameters.
        region (str): Region for which to retrieve default hyperparameters.
        include_container_hyperparameters (bool): True if container hyperparameters
            should be returned as well. Container hyperparameters are not used to tune
            the specific algorithm, but rather by SageMaker Training to setup
            the training container environment. For example, there is a container hyperparameter
            that indicates the entrypoint script to use. These hyperparameters may be required
            when creating a training job with boto3, however the ``Estimator`` classes
            should take care of adding container hyperparameters to the job. (Default: False).
    Returns:
        dict: the hyperparameters to use for the model.
    """

    if region is None:
        region = JUMPSTART_DEFAULT_REGION_NAME

    model_specs = jumpstart_accessors.JumpStartModelsAccessor.get_model_specs(
        region=region, model_id=model_id, version=model_version
    )

    default_hyperparameters: Dict[str, str] = {}
    for hyperparameter in model_specs.hyperparameters:
        if (
            include_container_hyperparameters and hyperparameter.scope == VariableScope.CONTAINER
        ) or hyperparameter.scope == VariableScope.ALGORITHM:
            default_hyperparameters[hyperparameter.name] = str(hyperparameter.default)
    return default_hyperparameters


def _retrieve_default_environment_variables(
    model_id: str,
    model_version: str,
    region: Optional[str],
):
    """Retrieves the inference environment variables for the model matching the given arguments.

    Args:
        model_id (str): JumpStart model ID of the JumpStart model for which to
            retrieve the default environment variables.
        model_version (str): Version of the JumpStart model for which to retrieve the
            default environment variables.
        region (Optional[str]): Region for which to retrieve default environment variables.

    Returns:
        dict: the inference environment variables to use for the model.
    """

    if region is None:
        region = JUMPSTART_DEFAULT_REGION_NAME

    model_specs = jumpstart_accessors.JumpStartModelsAccessor.get_model_specs(
        region=region, model_id=model_id, version=model_version
    )

    default_environment_variables: Dict[str, str] = {}
    for environment_variable in model_specs.inference_environment_variables:
        default_environment_variables[environment_variable.name] = str(environment_variable.default)
    return default_environment_variables



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Helper functions for model card."""
from __future__ import absolute_import

import logging
from abc import ABC, abstractmethod
from typing import Any, Optional, Union, List
import inspect
from enum import Enum
import json
import copy
import hashlib
import collections
from botocore.exceptions import ClientError
from boto3.session import Session


logger = logging.getLogger(__name__)


class _JSONEncoder(json.JSONEncoder):
    """Custom JSON encoder for model card"""

    def default(self, o):
        """Add additional condition to customize the encoding strategy for model card object.

        Args:
            o (Any): object.
        """
        if hasattr(o, "_to_request_dict"):
            # custom object with _to_request_dict method
            result = o._to_request_dict()
        else:
            # fall back to default encoding strategy
            result = json.JSONEncoder.default(self, o)

        return result


class _DefaultToRequestDict(object):
    """Provide a default behavior for to_request_dict method."""

    def _clean_descriptor_name(self, name: str):
        """Update the attribute name to be the same as schema.

        Args:
            name (str): attribute name.
        """
        if name.startswith("_") and hasattr(self, name[1:]):
            name = name[1:]

        return name

    def _to_request_dict(self):
        """Implement this method in a subclass to return a custom request_dict."""
        request_data = {}
        for attr, value in self.__dict__.items():
            if value is not None:
                name = self._clean_descriptor_name(attr)
                request_data[name] = value

        return request_data


class _DefaultFromDict(object):
    """Provide a default behavior for from_dict method."""

    @classmethod
    def _from_dict(cls, raw_data: dict):
        """Implement this method in a subclass to custom load JSON object.

        Args:
            raw_data (dict): model card json data ready to be encode.
        """
        args = copy.deepcopy(raw_data)

        return cls(**args)


class _DescriptorBase(ABC):
    """Base class for model card descriptor source: https://docs.python.org/3/howto/descriptor.html#complete-practical-example."""  # noqa E501  # pylint: disable=c0301

    def __set_name__(self, owner: type, name: str):
        """Set descriptor private name.

        Args:
            owner (type): attribute Parent class.
            name (str): attribute name.
        """
        self.private_name = "_" + name  # pylint: disable=W0201

    def __get__(self, obj, objtype: type = None):
        """Get the object attribute value during obj.attribute.

        Args:
            objtype (type): attribute Parent class (default: None).
        """
        return getattr(obj, self.private_name)

    def __set__(self, obj: object, value: Any):
        """Set the object attribute value during obj.attribute = value.

        Args:
            obj (object): Attribute object.
            value (Any): Value assigned to the attribute.
        """
        self.validate(value)
        if self.require_decode(value):
            # custom decode method
            value = self.decode(value)
        setattr(obj, self.private_name, value)

    @abstractmethod
    def validate(self, value):
        """Custom validators need to inherit from _DescriptorBase and must supply a validate method to test various restrictions as needed.

        Args:
            value (Any): value to be validated.
        """  # noqa E501  # pylint: disable=c0301
        pass  # pylint: disable=W0107

    @abstractmethod
    def require_decode(self, value: dict):
        """Decide if the object requires to run decode method. Implement this method in a subclass to return a custom request_dict.

        Args:
            value (dict): raw data to be decoded.
        """  # noqa E501  # pylint: disable=c0301
        pass  # pylint: disable=W0107

    @abstractmethod
    def decode(self, value: dict):
        """Decode routine of the JSON object.

        Args:
            value (dict): raw data to be decoded.
        """
        pass  # pylint: disable=W0107


class _OneOf(_DescriptorBase):
    """Verifies that a value is one of a restricted set of options"""

    def __init__(self, enumerator: Enum):
        """Initialize a OneOf descriptor.

        Args:
            options (Enum): options to be verified.
        """
        self.options = set(enumerator)
        self.enumerator = enumerator
        self.enumerator_reverse = {i.value: i.name for i in self.enumerator}

    def validate(self, value):
        """Check if value is valid for enumerator.

        Args:
            value (str or Enum): enumerator value.

        Raises:
            ValueError: value is not in the enumerator.
        """
        if value is not None and value not in self.options:
            if not isinstance(value, Enum):
                expect = sorted([i.value for i in self.options])
            else:
                expect = sorted([str(i) for i in self.options])

            raise ValueError(f"Expected {str(value)} to be one of {expect}")

    def require_decode(self, value: Union[Enum, str]):
        """Check if the value requires decoding.

        Args:
            value (Enum or str): raw data to be decoded to Enum.
        """
        if isinstance(value, (Enum, type(None))):
            return False
        return True

    def decode(self, value: str):
        """Decode the value to an enumerator.

        Args:
            value (str): raw data to be decoded to Enum.
        """
        return getattr(self.enumerator, self.enumerator_reverse[value])


class _IsList(_DescriptorBase):
    """List object."""

    def __init__(self, item_type: object, max_size: Optional[int] = None):
        """Initialize an IsList descriptor.

        Args:
            item_type (object): Item class in the list.
            max_size (int, optional): max size of the list. Defaults to None, i.e. no size limit for the list.
        """  # noqa E501  # pylint: disable=c0301
        self.item_type = item_type
        self.max_size = max_size

    def validate(self, value: List):
        """Check if value is valid for _MaxSizeList.

        Args:
            value (List)

        Raises:
            ValueError: value is not a list.
        """
        if value is not None and not isinstance(value, list):
            raise ValueError(f"Please assign a list to {self.private_name[1:]}")

    def require_decode(self, value: List):  # pylint: disable=w0613
        """Check if the value requires decoding.

        Args:
            value (List): raw data to be decoded to _MaxSizeList.
        """
        # required to convert the list to _MaxSizeList
        return True

    def decode(self, value: List):
        """Decode the value to a _MaxSizeList.

        Args:
            value (List): raw data to be decoded to _MaxSizeList.
        """
        # default value for _IsList attribute is []
        array = []
        if value is not None:
            for item in value:
                if isinstance(value[0], self.item_type):
                    array.append(item)
                else:
                    array.append(self.item_type._from_dict(item))

        return _MaxSizeArray(self.max_size, self.item_type, array)


class _IsModelCardObject(_DescriptorBase):
    """Model Card object class."""

    def __init__(self, custom_class: object):
        """Initialize a model card object descriptor.

        Args:
            custom_class (object): model card object class.
        """
        self.custom_class = custom_class

    def validate(self, value: Union[dict, object]):
        """Validate if data is value for model card object.

        Args:
            value (dict or object)

        Raises:
            ValueError: value is not a dict or custom_class.
        """
        from_dict = isinstance(value, dict)
        if value is not None and not from_dict and not isinstance(value, self.custom_class):
            raise ValueError(
                f"Expected {type(value)} instance to be of class {self.custom_class.__name__}."
            )  # noqa E501  # pylint: disable=c0301

    def require_decode(self, value: Union[dict, object]):
        """Check if value requires decoding.

        Args:
            value (dict or object): raw data to be decoded to custom_class object.
        """
        if isinstance(value, (self.custom_class, type(None))):
            res = False
        else:
            res = True
        return res

    def decode(self, value: dict):
        """Decode the value to a custom class object.

        Args:
            value (dict): raw data to be decoded to custom_class object.

        Raises:
            TypeError: Attributes in the data don't match the class definition.
        """
        try:
            return self.custom_class._from_dict(value)
        except TypeError as e:
            raise TypeError(f"class {self.custom_class} {str(e)}")


class _MaxSizeArray(collections.abc.MutableSequence):  # pylint: disable=too-many-ancestors
    """Array with maximum size and items of the same type."""

    def __init__(self, max_size: int, item_type: Any, array: List = None):
        """Initialize a Max Size Array.

        Args:
            max_size (int): array max size.
            item_type (Any): array item type.
            array (List, optional): initial array items (default: None).
        """
        super().__init__()
        if max_size is None:
            max_size = float("inf")
        if max_size < 0:
            raise ValueError("Max size has to be positive integer")
        self._max_size = max_size
        if not inspect.isclass(item_type):
            raise ValueError("Item type has to be a class")
        self._item_type = item_type
        self.list = []
        self._initialize_data(array)

    def __len__(self):
        """Return len(self)."""
        return len(self.list)

    def __getitem__(self, index):
        """Get Self[index]

        Args:
            index (int): List index.
        """
        return self.list[index]

    def __delitem__(self, index):
        """Delete self[index].

        Args:
            index (int): List index.
        """
        del self.list[index]

    def __setitem__(self, index: int, value: Any):
        """Set self[key] to value.

        Args:
            index (int): List index.
            value (Any): List element.
        """
        self.check(value)
        self.list[index] = value

    def __str__(self):
        """Return str(self)."""
        return str([i for i in self.list])  # pylint: disable=unnecessary-comprehension

    def __repr__(self):
        """Return repr(self)."""
        return self.__str__()

    def __eq__(self, other: Any):
        """Return self==value.

        Args:
            other (Any): The other object used for comparison.
        """
        return self.list == other

    def _initialize_data(self, array: List):
        """Initialize the max size list from a list.

        Args:
            array (list): initial list data.

        Raises:
            ValueError: size of the array is larger than max_size
        """
        if array:
            if len(array) > self._max_size:
                raise ValueError(
                    f"Data size {len(array)} exceed the maximum size of {self._max_size}"
                )
            for item in array:
                self.append(item)

    def check(self, value: Any):
        """Check if the item is valid.

        Args:
            value (Any): item in the max size list.

        Raises:
            ValueError: Item's type is not the same as item_type.
            ValueError: Result list size is larger than max_size.
        """

        if not isinstance(value, self._item_type):
            raise TypeError(
                f"Provided item type is {type(value)} and Expected the item type is {self._item_type}"  # noqa E501  # pylint: disable=c0301
            )
        if len(self.list) >= self._max_size:
            raise ValueError(f"Exceed the maximum size of {self._max_size}")

    def insert(self, index, value):
        """Insert item into the list. (implicitly overwrite append method)"""
        self.check(value)
        self.list.insert(index, value)

    def _to_request_dict(self):
        """Create data for request body"""
        return list(self)

    def to_map(self, key_attribute: str):
        """Generate a map from the list items for fast look up

        Args:
            key_attribute (str): the attribute in the item where the map key is coming from
        """
        map_ = {}
        for item in self.list:
            key = getattr(item, key_attribute, None)
            if key:
                map_[key] = item

        return map_


def _hash_content_str(content: str):
    """Create hash for content string

    Args:
        content (str): content string from json.dumps.
    """
    dhash = hashlib.md5()
    encoded = json.dumps(json.loads(content), sort_keys=True, default=str).encode()
    dhash.update(encoded)

    return dhash.hexdigest()


def _read_s3_json(session: Session, bucket: str, key: str):
    """Read json file from S3 bucket.

    Args:
        session (Session): boto3 session.
        bucket (str): S3 bucket name.
        key (str): S3 key of the json file.
    """
    client = session.client("s3")
    try:
        data = client.get_object(Bucket=bucket, Key=key)
    except ClientError as ex:
        if ex.response["Error"]["Code"] == "NoSchKey":  # pylint: disable=r1705
            logger.warning("Metric file %s does not exist in %s.", key, bucket)
            return {}
        else:
            raise

    result = {}
    if data["ContentType"] == "application/json":
        result = json.loads(data["Body"].read().decode("utf-8"))
    else:
        logger.warning("Invalid file type %s. application/json is expected.", data["ContentType"])

    return result



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""The properties definitions for workflow."""
from __future__ import absolute_import

from abc import ABCMeta
from typing import Dict, Union, List

import attr

import botocore.loaders

from sagemaker.workflow.entities import Expression, PipelineVariable


class PropertiesMeta(ABCMeta):
    """Load an internal shapes attribute from the botocore service model

    for sagemaker and emr service.
    """

    _shapes_map = dict()
    _primitive_types = {"string", "boolean", "integer", "float"}

    def __new__(mcs, *args, **kwargs):
        """Loads up the shapes from the botocore service model."""
        if len(mcs._shapes_map.keys()) == 0:
            loader = botocore.loaders.Loader()

            sagemaker_model = loader.load_service_model("sagemaker", "service-2")
            emr_model = loader.load_service_model("emr", "service-2")
            mcs._shapes_map["sagemaker"] = sagemaker_model["shapes"]
            mcs._shapes_map["emr"] = emr_model["shapes"]

        return super().__new__(mcs, *args, **kwargs)


class Properties(PipelineVariable, metaclass=PropertiesMeta):
    """Properties for use in workflow expressions."""

    def __init__(
        self,
        step_name: str,
        path: str = None,
        shape_name: str = None,
        shape_names: List[str] = None,
        service_name: str = "sagemaker",
    ):
        """Create a Properties instance representing the given shape.

        Args:
            step_name (str): The name of the Step this Property belongs to.
            path (str): The relative path of this Property value.
            shape_name (str): The botocore service model shape name.
            shape_names (str): A List of the botocore service model shape name.
        """
        self.step_name = step_name
        self.path = path

        shape_names = [] if shape_names is None else shape_names
        self._shape_names = shape_names if shape_name is None else [shape_name] + shape_names

        shapes = Properties._shapes_map.get(service_name, {})

        for name in self._shape_names:
            shape = shapes.get(name, {})
            shape_type = shape.get("type")
            if shape_type in Properties._primitive_types:
                self.__str__ = name
            elif shape_type == "structure":
                members = shape["members"]
                for key, info in members.items():
                    if shapes.get(info["shape"], {}).get("type") == "list":
                        self.__dict__[key] = PropertiesList(
                            step_name=step_name,
                            path=".".join(filter(None, (path, key))),
                            shape_name=info["shape"],
                            service_name=service_name,
                        )
                    elif shapes.get(info["shape"], {}).get("type") == "map":
                        self.__dict__[key] = PropertiesMap(
                            step_name=step_name,
                            path=".".join(filter(None, (path, key))),
                            shape_name=info["shape"],
                            service_name=service_name,
                        )
                    else:
                        self.__dict__[key] = Properties(
                            step_name=step_name,
                            path=".".join(filter(None, (path, key))),
                            shape_name=info["shape"],
                            service_name=service_name,
                        )

    @property
    def expr(self):
        """The 'Get' expression dict for a `Properties`."""
        prefix = f"Steps.{self.step_name}"
        full_path = prefix if self.path is None else f"{prefix}.{self.path}"
        return {"Get": full_path}

    @property
    def _referenced_steps(self) -> List[str]:
        """List of step names that this function depends on."""
        return [self.step_name]


class PropertiesList(Properties):
    """PropertiesList for use in workflow expressions."""

    def __init__(
        self, step_name: str, path: str, shape_name: str = None, service_name: str = "sagemaker"
    ):
        """Create a PropertiesList instance representing the given shape.

        Args:
            step_name (str): The name of the Step this Property belongs to.
            path (str): The relative path of this Property value.
            shape_name (str): The botocore service model shape name.
            service_name (str): The botocore service name.
        """
        super(PropertiesList, self).__init__(step_name, path, shape_name)
        self.shape_name = shape_name
        self.service_name = service_name
        self._items: Dict[Union[int, str], Properties] = dict()

    def __getitem__(self, item: Union[int, str]):
        """Populate the indexing item with a Property, for both lists and dictionaries.

        Args:
            item (Union[int, str]): The index of the item in sequence.
        """
        if item not in self._items.keys():
            shape = Properties._shapes_map.get(self.service_name, {}).get(self.shape_name)
            member = shape["member"]["shape"]
            if isinstance(item, str):
                property_item = Properties(self.step_name, f"{self.path}['{item}']", member)
            else:
                property_item = Properties(self.step_name, f"{self.path}[{item}]", member)
            self._items[item] = property_item

        return self._items.get(item)


class PropertiesMap(Properties):
    """PropertiesMap for use in workflow expressions."""

    def __init__(
        self, step_name: str, path: str, shape_name: str = None, service_name: str = "sagemaker"
    ):
        """Create a PropertiesMap instance representing the given shape.

        Args:
            step_name (str): The name of the Step this Property belongs to.
            path (str): The relative path of this Property value.
            shape_name (str): The botocore service model shape name.
            service_name (str): The botocore service name.
        """
        super(PropertiesMap, self).__init__(step_name, path, shape_name)
        self.shape_name = shape_name
        self.service_name = service_name
        self._items: Dict[Union[int, str], Properties] = dict()

    def __getitem__(self, item: Union[int, str]):
        """Populate the indexing item with a Property, for both lists and dictionaries.

        Args:
            item (Union[int, str]): The index of the item in sequence.
        """
        if item not in self._items.keys():
            shape = Properties._shapes_map.get(self.service_name, {}).get(self.shape_name)
            member = shape["value"]["shape"]
            if isinstance(item, str):
                property_item = Properties(self.step_name, f"{self.path}['{item}']", member)
            else:
                property_item = Properties(self.step_name, f"{self.path}[{item}]", member)
            self._items[item] = property_item

        return self._items.get(item)


@attr.s
class PropertyFile(Expression):
    """Provides a property file struct.

    Attributes:
        name (str): The name of the property file for reference with `JsonGet` functions.
        output_name (str): The name of the processing job output channel.
        path (str): The path to the file at the output channel location.
    """

    name: str = attr.ib()
    output_name: str = attr.ib()
    path: str = attr.ib()

    @property
    def expr(self) -> Dict[str, str]:
        """The expression dict for a `PropertyFile`."""
        return {
            "PropertyFileName": self.name,
            "OutputName": self.output_name,
            "FilePath": self.path,
        }



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"). You
# may not use this file except in compliance with the License. A copy of
# the License is located at
#
#     http://aws.amazon.com/apache2.0/
#
# or in the "license" file accompanying this file. This file is
# distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF
# ANY KIND, either express or implied. See the License for the specific
# language governing permissions and limitations under the License.
"""Contains the helper classes for SageMaker Experiment."""
from __future__ import absolute_import

import json
import logging
import os

import botocore

from sagemaker.experiments._utils import is_already_exist_error

logger = logging.getLogger(__name__)


_DEFAULT_ARTIFACT_PREFIX = "trial-component-artifacts"
_DEFAULT_ARTIFACT_TYPE = "Tracker"


class _ArtifactUploader(object):
    """Artifact uploader"""

    def __init__(
        self,
        trial_component_name,
        sagemaker_session,
        artifact_bucket=None,
        artifact_prefix=_DEFAULT_ARTIFACT_PREFIX,
    ):
        """Initialize a `_ArtifactUploader` instance.

        Args:
            trial_component_name (str): The name of the trial component,
                which is used to generate the S3 path to upload the artifact to.
            sagemaker_session (sagemaker.session.Session): Session object which
                manages interactions with Amazon SageMaker APIs and any other
                AWS services needed.
            artifact_bucket (str): The S3 bucket to upload the artifact to.
                If not specified, the default bucket defined in `sagemaker_session`
                will be used.
            artifact_prefix (str): The S3 key prefix used to generate the S3 path
                to upload the artifact to (default: "trial-component-artifacts").
        """
        self.sagemaker_session = sagemaker_session
        self.trial_component_name = trial_component_name
        self.artifact_bucket = artifact_bucket
        self.artifact_prefix = artifact_prefix
        self._s3_client = self.sagemaker_session.boto_session.client("s3")

    def upload_artifact(self, file_path):
        """Upload an artifact file to S3.

        Args:
            file_path (str): the file path of the artifact

        Returns:
            (str, str): The s3 URI of the uploaded file and the etag of the file.

        Raises:
            ValueError: If file does not exist.
        """
        file_path = os.path.expanduser(file_path)
        if not os.path.isfile(file_path):
            raise ValueError(
                "{} does not exist or is not a file. Please supply a file path.".format(file_path)
            )
        if not self.artifact_bucket:
            self.artifact_bucket = self.sagemaker_session.default_bucket()
        artifact_name = os.path.basename(file_path)
        artifact_s3_key = "{}/{}/{}".format(
            self.artifact_prefix, self.trial_component_name, artifact_name
        )
        self._s3_client.upload_file(file_path, self.artifact_bucket, artifact_s3_key)
        etag = self._try_get_etag(artifact_s3_key)
        return "s3://{}/{}".format(self.artifact_bucket, artifact_s3_key), etag

    def upload_object_artifact(self, artifact_name, artifact_object, file_extension=None):
        """Upload an artifact object to S3.

        Args:
            artifact_name (str): the name of the artifact.
            artifact_object (obj): the object of the artifact
            file_extension (str): Optional file extension.

        Returns:
            str: The s3 URI of the uploaded file and the version of the file.
        """
        if not self.artifact_bucket:
            self.artifact_bucket = self.sagemaker_session.default_bucket()
        if file_extension:
            artifact_name = (
                artifact_name + ("" if file_extension.startswith(".") else ".") + file_extension
            )
        artifact_s3_key = "{}/{}/{}".format(
            self.artifact_prefix, self.trial_component_name, artifact_name
        )
        self._s3_client.put_object(
            Body=json.dumps(artifact_object), Bucket=self.artifact_bucket, Key=artifact_s3_key
        )
        etag = self._try_get_etag(artifact_s3_key)
        return "s3://{}/{}".format(self.artifact_bucket, artifact_s3_key), etag

    def _try_get_etag(self, key):
        """Get ETag of given key and return None if not allowed

        Args:
            key (str): The S3 object key.

        Returns:
            str: The S3 object ETag if it allows, otherwise return None.
        """
        try:
            response = self._s3_client.head_object(Bucket=self.artifact_bucket, Key=key)
            return response["ETag"]
        except botocore.exceptions.ClientError as error:
            # requires read permissions
            logger.warning("Failed to get ETag of %s due to %s", key, error)
        return None


class _LineageArtifactManager(object):
    """A helper class to manage Lineage Artifacts"""

    def __init__(
        self,
        name,
        source_uri,
        etag,
        source_arn=None,
        dest_arn=None,
        artifact_type=_DEFAULT_ARTIFACT_TYPE,
    ):
        """Initialize a `_LineageArtifactManager` instance.

        Args:
            name (str): The name of the Lineage artifact to be created.
            source_uri (str): The source URI used to create the Lineage artifact.
            etag (str): The S3 Etag used to create the Lineage artifact.
            source_arn (str): The source ARN of a trail component to associate
                this Lineage artifact with (default: None).
            dest_arn (str): The destination ARN of a trial component to associate
                this Lineage artifact with (default: None).
            artifact_type (str): The type of the Lineage artifact (default: "Tracker").
        """
        self.name = name
        self.source_uri = source_uri
        self.etag = etag
        self.source_arn = source_arn
        self.dest_arn = dest_arn
        self.artifact_arn = None
        self.artifact_type = artifact_type

    def create_artifact(self, sagemaker_session):
        """Create the artifact by calling `CreateArtifact` API

        Args:
            sagemaker_session (sagemaker.session.Session): Session object which
                manages interactions with Amazon SageMaker APIs and any other
                AWS services needed.
        """
        source_ids = []
        if self.etag:
            source_ids.append({"SourceIdType": "S3ETag", "Value": self.etag})

        try:
            response = sagemaker_session.sagemaker_client.create_artifact(
                ArtifactName=self.name,
                ArtifactType=self.artifact_type,
                Source={"SourceUri": self.source_uri, "SourceTypes": source_ids},
            )
            self.artifact_arn = response["ArtifactArn"]
        except botocore.exceptions.ClientError as err:
            err_info = err.response["Error"]
            if not is_already_exist_error(err_info):
                raise
            logger.warning(
                "Skip creating the artifact since it already exists: %s", err_info["Message"]
            )

    def add_association(self, sagemaker_session):
        """Associate the artifact with a source/destination ARN (e.g. trial component arn)

        Args:
            sagemaker_session (sagemaker.session.Session): Session object which
                manages interactions with Amazon SageMaker APIs and any other
                AWS services needed.
        """
        source_arn = self.source_arn if self.source_arn else self.artifact_arn
        dest_arn = self.dest_arn if self.dest_arn else self.artifact_arn
        # if the trial component (job) is the source then it produced the artifact,
        # otherwise the artifact contributed to the trial component (job)
        association_edge_type = "Produced" if self.source_arn else "ContributedTo"
        try:
            sagemaker_session.sagemaker_client.add_association(
                SourceArn=source_arn, DestinationArn=dest_arn, AssociationType=association_edge_type
            )
        except botocore.exceptions.ClientError as err:
            err_info = err.response["Error"]
            if not is_already_exist_error(err_info):
                raise
            logger.warning(
                "Skip associating since the association already exists: %s", err_info["Message"]
            )


class _LineageArtifactTracker(object):
    """Lineage Artifact Tracker"""

    def __init__(self, trial_component_arn, sagemaker_session):
        """Initialize a `_LineageArtifactTracker` instance.

        Args:
            trial_component_arn (str): The ARN of the trial component to be
                associated with the input/output artifacts.
            sagemaker_session (sagemaker.session.Session): Session object which
                manages interactions with Amazon SageMaker APIs and any other
                AWS services needed.
        """
        self.trial_component_arn = trial_component_arn
        self.sagemaker_session = sagemaker_session
        self.artifacts = []

    def add_input_artifact(self, name, source_uri, etag, artifact_type):
        """Add a Lineage input artifact locally

        Args:
            name (str): The name of the Lineage input artifact to be added.
            source_uri (str): The source URI used to create the Lineage input artifact.
            etag (str): The S3 Etag used to create the Lineage input artifact.
            artifact_type (str): The type of the Lineage input artifact.
        """
        artifact = _LineageArtifactManager(
            name, source_uri, etag, dest_arn=self.trial_component_arn, artifact_type=artifact_type
        )
        self.artifacts.append(artifact)

    def add_output_artifact(self, name, source_uri, etag, artifact_type):
        """Add a Lineage output artifact locally

        Args:
            name (str): The name of the Lineage output artifact to be added.
            source_uri (str): The source URI used to create the Lineage output artifact.
            etag (str): The S3 Etag used to create the Lineage output artifact.
            artifact_type (str): The type of the Lineage output artifact.
        """
        artifact = _LineageArtifactManager(
            name, source_uri, etag, source_arn=self.trial_component_arn, artifact_type=artifact_type
        )
        self.artifacts.append(artifact)

    def save(self):
        """Persist any artifact data saved locally"""
        for artifact in self.artifacts:
            artifact.create_artifact(self.sagemaker_session)
            artifact.add_association(self.sagemaker_session)
